# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import subprocess
import uuid
from pathlib import Path
from typing import Any

from dedal.configuration.SpackConfig import SpackConfig
from dedal.enum.SpackConfigCommand import SpackConfigCommand
from dedal.error_handling.exceptions import SpackInstallPackagesException, SpackConcertizeException
from dedal.logger.logger_builder import get_logger
from dedal.serialize.PickleSerializer import PickleSerializer
from dedal.spack_factory.SpackCacheOperation import SpackCacheOperation
from dedal.spack_factory.SpackEnvOperation import SpackEnvOperation
from dedal.spack_factory.SpackToolOperation import SpackToolOperation
from dedal.tests.testing_variables import SPACK_VERSION
from dedal.utils.utils import run_command, git_clone_repo, log_command, set_bashrc_variable
from dedal.wrapper.spack_wrapper import check_spack_env


class SpackOperation:
    """
    This class should implement the methods necessary for installing spack, set up an environment, concretize and install packages.
    Factory design pattern is used because there are 2 cases: creating an environment from scratch or creating an environment from the buildcache.

    Attributes:
    -----------
    env : SpackDescriptor
        spack environment details
    repos : list[SpackDescriptor]
    upstream_instance : str
        path to Spack instance to use as upstream (optional)
    """

    def __init__(self, spack_config: SpackConfig = SpackConfig(), logger=get_logger(__name__)):
        self.spack_config = spack_config
        self.logger = logger
        self.spack_config.install_dir = spack_config.install_dir
        os.makedirs(self.spack_config.install_dir, exist_ok=True)
        if self.spack_config.spack_dir is None:
            self.spack_config.spack_dir = self.spack_config.install_dir / 'spack'
        self.env_path = None
        self.spack_setup_script = "" if self.spack_config.use_spack_global else f"source {self.spack_config.spack_dir / 'share' / 'spack' / 'setup-env.sh'}"
        self.spack_config.concretization_dir = spack_config.concretization_dir
        if self.spack_config.concretization_dir:
            os.makedirs(self.spack_config.concretization_dir, exist_ok=True)
        self.spack_config.buildcache_dir = spack_config.buildcache_dir
        if self.spack_config.buildcache_dir:
            os.makedirs(self.spack_config.buildcache_dir, exist_ok=True)
        if self.spack_config.env and spack_config.env.name:
            self.env_path: Path = spack_config.env.path / spack_config.env.name
            if self.spack_setup_script != "":
                self.spack_command_on_env = f'{self.spack_setup_script} && spack env activate -p {spack_config.view.value} {self.env_path}'
            else:
                self.spack_command_on_env = f'spack env activate -p {spack_config.view.value} {self.env_path}'
        else:
            self.spack_command_on_env = self.spack_setup_script
        if self.spack_config.env and spack_config.env.path:
            self.spack_config.env.path.mkdir(parents=True, exist_ok=True)
        self.spack_tool_operation = SpackToolOperation(spack_config=spack_config,
                                                       spack_setup_script=self.spack_setup_script,
                                                       spack_dir=self.spack_config.spack_dir)
        self.spack_env_operation = SpackEnvOperation(spack_config=spack_config,
                                                     spack_setup_script=self.spack_setup_script,
                                                     env_path=self.env_path,
                                                     spack_command_on_env=self.spack_command_on_env)
        self.spack_cache_operation = SpackCacheOperation(spack_config=spack_config,
                                                         spack_setup_script=self.spack_setup_script,
                                                         spack_command_on_env=self.spack_command_on_env,
                                                         spack_dir=self.spack_config.spack_dir)
        self.serializer = PickleSerializer(file_path=self.spack_config.install_dir,
                                           file_name=self.spack_config.serialize_name)

    def setup_spack_env(self):
        """
        This method prepares a spack environment by fetching/creating the spack environment and adding the necessary repos
        """
        if self.spack_config.system_name:
            set_bashrc_variable('SYSTEMNAME', self.spack_config.system_name, self.spack_config.bashrc_path,
                                logger=self.logger)
            os.environ['SYSTEMNAME'] = self.spack_config.system_name
        if self.spack_config.spack_dir.exists() and self.spack_config.spack_dir.is_dir():
            set_bashrc_variable('SPACK_USER_CACHE_PATH', str(self.spack_config.spack_dir / ".spack"),
                                self.spack_config.bashrc_path,

                                logger=self.logger)
            set_bashrc_variable('SPACK_USER_CONFIG_PATH', str(self.spack_config.spack_dir / ".spack"),
                                self.spack_config.bashrc_path,
                                logger=self.logger)
            self.logger.debug('Added env variables SPACK_USER_CACHE_PATH and SPACK_USER_CONFIG_PATH')
        else:
            self.logger.error(f'Invalid installation path: {self.spack_config.spack_dir}')
        # Restart the bash after adding environment variables
        if self.spack_config.env:
            self.spack_env_operation.setup_spack_environment()
        if self.spack_config.install_dir.exists():
            for repo in self.spack_config.repos:
                repo_dir = self.spack_config.install_dir / repo.path / repo.name
                git_clone_repo(repo.name, repo_dir, repo.git_path, logger=self.logger, git_branch=repo.git_branch)
                if not self.spack_repo_exists(repo.name):
                    self.add_spack_repo(repo.path, repo.name)
                    self.logger.debug(f'Added spack repository {repo.name}')
                else:
                    self.logger.debug(f'Spack repository {repo.name} already added')

    @check_spack_env
    def concretize_spack_env(self, force=True, fresh=False, test=None):
        """Concretization step for a spack environment
            Args:
                force (bool): TOverrides an existing concretization when set to True
                test: which test dependencies should be included
            Raises:
                NoSpackEnvironmentException: If the spack environment is not set up.
        """
        force = '--force' if force else ''
        fresh = '--fresh' if fresh else ''
        test = f'--test {test}' if test else ''
        run_command("bash", "-c",
                    f'{self.spack_command_on_env} && spack concretize {fresh} {force} {test}',
                    check=True,
                    logger=self.logger,
                    info_msg=f'Concertization step for {self.spack_config.env.name}',
                    exception_msg=f'Failed the concertization step for {self.spack_config.env.name}',
                    exception=SpackConcertizeException)

    @check_spack_env
    def install_packages(self, jobs: int, signed=True, fresh=False, debug=False, test=None):
        """Installs all spack packages.
        Raises:
            NoSpackEnvironmentException: If the spack environment is not set up.
        """
        signed = '' if signed else '--no-check-signature'
        fresh = '--fresh' if fresh else ''
        debug = '--debug' if debug else ''
        test = f'--test {test}' if test else ''
        install_result = run_command("bash", "-c",
                                     f'{self.spack_command_on_env} && spack {debug} install {signed} -j {jobs} {fresh} {test}',
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE,
                                     text=True,
                                     logger=self.logger,
                                     info_msg=f"Installing spack packages for {self.spack_config.env.name}",
                                     exception_msg=f"Error installing spack packages for {self.spack_config.env.name}",
                                     exception=SpackInstallPackagesException)
        log_command(install_result, str(Path(os.getcwd()).resolve() / ".generate_cache.log"))
        if install_result.returncode == 0:
            self.logger.info(f'Finished installation of spack packages from scratch')
        else:
            self.logger.error(f'Something went wrong during installation. Please check the logs.')
        return install_result

    def create_fetch_spack_environment(self):
        self.spack_env_operation.setup_spack_environment()

    def add_spack_repo(self, repo_path: Path, repo_name: str):
        self.spack_env_operation.add_spack_repo(repo_path, repo_name)

    def spack_repo_exists(self, repo_name: str) -> bool | None:
        return self.spack_env_operation.spack_repo_exists(repo_name)

    def spack_env_exists(self):
        return self.spack_env_operation.spack_env_exists()

    @check_spack_env
    def get_compiler_version(self):
        return self.spack_env_operation.get_compiler_version()

    def get_spack_installed_version(self):
        return self.spack_tool_operation.get_spack_installed_version()

    def reindex(self):
        self.spack_cache_operation.reindex()

    def spec_pacakge(self, package_name: str, full_format=False):
        return self.spack_env_operation.spec_pacakge(package_name, full_format)

    def create_gpg_keys(self):
        self.spack_cache_operation.create_gpg_keys()

    def add_mirror(self, mirror_name: str, mirror_path: Path, signed=False, autopush=False, global_mirror=False):
        return self.spack_cache_operation.add_mirror(mirror_name, mirror_path, signed, autopush, global_mirror)

    @check_spack_env
    def trust_gpg_key(self, public_key_path: str):
        self.spack_cache_operation.trust_gpg_key(public_key_path)

    def config(self, config_type: SpackConfigCommand, config_parameter):
        self.spack_env_operation.config(config_type, config_parameter)

    def mirror_list(self):
        return self.spack_cache_operation.mirror_list()

    def remove_mirror(self, mirror_name: str):
        self.spack_cache_operation.remove_mirror(mirror_name)

    def update_buildcache_index(self, mirror_path: str):
        self.spack_cache_operation.update_buildcache_index(mirror_path)

    def install_gpg_keys(self):
        self.spack_cache_operation.install_gpg_keys()

    @check_spack_env
    def find_packages(self):
        return self.spack_env_operation.find_packages()

    def find_package(self, package, long: bool, variants: bool, paths: bool):
        return self.spack_env_operation.find_package(package, long, variants, paths)

    def fetch(self, dependencies=False, missing=False):
        return self.spack_env_operation.fetch(dependencies, missing)

    def install_spack(self, spack_version=f'{SPACK_VERSION}', spack_repo='https://github.com/spack/spack',
                      bashrc_path=os.path.expanduser("~/.bashrc")):
        self.spack_tool_operation.install_spack(spack_version, spack_repo, bashrc_path)

    def check_installed_spack_packages(self, env_path: Path):
        return self.spack_env_operation.check_installed_spack_packages(env_path)

    def create_build_cache(self, package=None, unsigned=True, only=False):
        return self.spack_cache_operation.create_build_cache(package, unsigned, only)

    def serialize(self) -> None:
        """
        Serialize spack configuration to the given file path.
        """
        self.serializer.serialize(self)

    @classmethod
    def deserialize(cls, file_location: Path, file_name: str) -> Any:
        return PickleSerializer(file_location.resolve(), file_name).deserialize()

    def merge_envs(self, env_path: Path, site_config_path: Path):
        return self.spack_env_operation.merge_envs(env_path, site_config_path)

    def create_load_env_script(self):
        self.spack_env_operation.create_load_env_script()

    def add_package(self, package: str):
        self.spack_env_operation.add_package(package)

    def remove_package(self, package: str, force: bool = False):
        self.spack_env_operation.remove_package(package, force)

    def spack_clean(self, bootstrap: bool = False):
        self.spack_env_operation.spack_clean(bootstrap)



# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import glob
import os
from pathlib import Path
from dedal.enum.SpackConfigCommand import SpackConfigCommand
from dedal.error_handling.exceptions import MissingAttributeException
from dedal.spack_factory.SpackEnvOperation import SpackEnvOperation
from dedal.tests.testing_variables import SPACK_VERSION
from dedal.utils.utils import run_command
from dedal.logger.logger_builder import get_logger
from dedal.configuration.SpackConfig import SpackConfig


class SpackToolOperation:
    def __init__(self, spack_config: SpackConfig = SpackConfig(), logger=get_logger(__name__), spack_setup_script=None,
                 spack_dir=Path('./').resolve()):
        self.spack_config = spack_config
        self.logger = logger
        if spack_setup_script:
            self.spack_setup_script = spack_setup_script
        else:
            raise MissingAttributeException(f'Missing attribute for class {__name__}')
        self.spack_dir = spack_dir

    def get_spack_installed_version(self):
        """Returns the spack installed version"""
        spack_version = run_command("bash", "-c", f'{self.spack_setup_script} && spack --version',
                                    capture_output=True, text=True, check=True,
                                    logger=self.logger,
                                    info_msg=f"Getting spack version",
                                    exception_msg=f"Error retrieving Spack version")
        if spack_version:
            return spack_version.stdout.strip().split()[0]
        return None

    def install_spack(self, spack_version=f'{SPACK_VERSION}', spack_repo='https://github.com/spack/spack',
                      bashrc_path=os.path.expanduser("~/.bashrc")):
        """Install spack.
            Args:
                spack_version (str): spack version
                spack_repo (str): Git path to the Spack repository.
                bashrc_path (str): Path to the .bashrc file.
        """
        spack_version = f'v{spack_version}'
        try:
            user = os.getlogin()
        except OSError:
            user = None

        self.logger.info(f"Starting to install Spack into {self.spack_dir} from branch {spack_version}")
        if not self.spack_dir.exists():
            run_command(
                "git", "clone", "--depth", "1",
                "-c", "advice.detachedHead=false",
                "-c", "feature.manyFiles=true",
                "--branch", spack_version, spack_repo, self.spack_dir
                , check=True, logger=self.logger)
            self.logger.debug("Cloned spack")
        else:
            self.logger.debug("Spack already cloned.")

        if bashrc_path:
            # ensure the file exists before opening it
            if not os.path.exists(bashrc_path):
                open(bashrc_path, "w").close()
            # add spack setup commands to .bashrc
            with open(bashrc_path, "a") as bashrc:
                bashrc.write(f'export PATH="{self.spack_dir}/bin:$PATH"\n')
                spack_setup_script = f"source {self.spack_dir / 'share' / 'spack' / 'setup-env.sh'}"
                bashrc.write(f"{spack_setup_script}\n")
            self.logger.info("Added Spack PATH to .bashrc")
        if user:
            run_command("chown", "-R", f"{user}:{user}", self.spack_dir, check=True, logger=self.logger,
                        info_msg='Adding permissions to the logged in user')
        self.logger.info("Spack install completed")
        if self.spack_config.use_spack_global is True and bashrc_path is not None:
            # Restart the bash only of the spack is used globally
            self.logger.info('Restarting bash')
            run_command("bash", "-c", f"source {bashrc_path}", check=True, logger=self.logger, info_msg='Restart bash')
            os.system("exec bash")
        # Configure upstream Spack instance if specified
        if self.spack_config.upstream_instance:
            search_path = os.path.join(self.spack_config.upstream_instance, 'spack', 'opt', 'spack', '**', '.spack-db')
            spack_db_dirs = glob.glob(search_path, recursive=True)
            upstream_prefix = [os.path.dirname(dir) for dir in spack_db_dirs]
            spack_env_operation = SpackEnvOperation(spack_config=self.spack_config,
                                                    spack_setup_script=self.spack_setup_script)
            for prefix in upstream_prefix:
                # todo fix
                spack_env_operation.config(SpackConfigCommand.ADD, f':upstream-spack-instance:install_tree:{prefix}')
            self.logger.info("Added upstream spack instance")



# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import sys
from pathlib import Path
from typing import List

from dedal.build_cache.BuildCacheManagerOciDefault import BuildCacheManagerOciDefault
from dedal.logger.logger_builder import get_logger

default_remote_cache = (("HARBOR_HOST" in os.environ) and ("HARBOR_PROJECT" in os.environ)) and (
    "{}/{}/source_cache".format(os.environ["HARBOR_HOST"], os.environ["HARBOR_PROJECT"])) or ""
default_remote_cache_type = 'oci'
default_remote_cache_username = "HARBOR_USERNAME" in os.environ and Path(os.environ["HARBOR_USERNAME"]) or ""
default_remote_cache_password = "HARBOR_PASSWORD" in os.environ and Path(os.environ["HARBOR_PASSWORD"]) or None
default_local_cache = "YASHCHIKI_CACHE_SOURCE" in os.environ and Path(
    os.environ["YASHCHIKI_CACHE_SOURCE"]) or os.path.expanduser("~/.dedal/cache/")


# migration of /cli/update_cached_sources.py
class CachedSourceUploader:
    def __init__(
            self,
            path_missing: Path,
            remote_cache: str = default_remote_cache,
            remote_cache_type: str = default_remote_cache_type,
            remote_cache_username: str = default_remote_cache_username,
            remote_cache_password: str = default_remote_cache_password,
            local_cache: str = default_local_cache
    ):
        self.path_missing = path_missing
        self.remote_cache = remote_cache
        self.remote_cache_type = remote_cache_type
        self.remote_cache_username = remote_cache_username
        self.remote_cache_password = remote_cache_password
        self.local_cache = Path(local_cache).resolve()
        self.logger = get_logger(__name__)

        self.oci = None
        if remote_cache_type == "oci":
            parts = self.remote_cache.split("/")
            if len(parts) < 3:
                raise ValueError("Remote cache must be in the format <host>/<project>/<version>")
            registry_host = parts[0]
            registry_project = parts[1]
            cache_version = parts[2] if len(parts) > 2 else 'source_cache'
            self.oci = BuildCacheManagerOciDefault(
                registry_host=registry_host,
                registry_project=registry_project,
                registry_username=self.remote_cache_username,
                registry_password=self.remote_cache_password,
                cache_version=cache_version,
            )
            if self.oci is None:
                raise RuntimeError("Failed to initialize BuildCacheManagerOci")

    def upload_missing_sources(self):
        if not self.path_missing.exists():
            self.logger.info(f"File w/ missing cached source information is not available: {self.path_missing}")
            sys.exit(0)

        with open(self.path_missing, "r") as fd:
            missing_file_paths: List[str] = fd.readlines()

        for path in missing_file_paths:
            stripped_path = path.strip()
            full_path = self.local_cache / stripped_path
            if not full_path.exists():
                self.logger.info(f"Missing file not found in local cache: {full_path}")
                continue

            tag = os.path.basename(stripped_path)
            rel_path = os.path.dirname(stripped_path)
            try:
                self.oci.client.push(
                    files=[str(full_path)],
                    target=f"{self.remote_cache}:{tag}",
                    manifest_annotations={"path": rel_path},
                    disable_path_validation=True
                )
                self.logger.info(f"Successfully uploaded '{stripped_path}' to '{self.remote_cache}:{tag}'")
            except Exception as e:
                self.logger.info(f"Uploading of '{stripped_path}' to '{self.remote_cache}:{tag}' failed. Reason: {e}")



# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
from pathlib import Path


class SpackDescriptor:
    """"
    Provides details about the spack environment
    """

    def __init__(self, name: str, path: Path = Path(os.getcwd()).resolve(), git_path: str = None, git_branch: str = ''):
        self.name = name
        self.path = path.resolve() if isinstance(path, Path) else Path(path).resolve()
        self.git_path = git_path
        self.git_branch = git_branch



# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re
import os
import shutil
import subprocess
import sys
from pathlib import Path

from dedal.utils.spack_utils import extract_spack_packages

from dedal.enum.SpackConfigCommand import SpackConfigCommand
from dedal.error_handling.exceptions import BashCommandException, MissingAttributeException, \
    NoSpackEnvironmentException, SpackRepoException, SpackConfigException, SpackFindException, SpackSpecException, \
    SpackFetchException, SpackMergeEnvsException, CreateLoadEnvException, SpackAddException, SpackRemoveException
from dedal.utils.utils import git_clone_repo, run_command, count_files_in_folder
from dedal.logger.logger_builder import get_logger
from dedal.configuration.SpackConfig import SpackConfig


class SpackEnvOperation:
    def __init__(self, spack_config: SpackConfig = SpackConfig(), logger=get_logger(__name__), spack_setup_script=None,
                 env_path=None, spack_command_on_env=None):
        self.spack_config = spack_config
        self.logger = logger
        self.env_path = env_path
        if spack_setup_script and spack_command_on_env:
            self.spack_setup_script = spack_setup_script
            self.spack_command_on_env = spack_command_on_env
        else:
            raise MissingAttributeException(f'Missing attribute for class {__name__}')

    def setup_spack_environment(self):
        """
        Fetches a spack environment if the git path is defined, otherwise creates it. If site-config is defined, it merges the spack.yaml files
        """
        site_config_path = self.env_path / 'site-config'
        if self.spack_config.env and self.spack_config.env.git_path:
            self.logger.info('Setup spack environment: fetch environment from git repo')
            git_clone_repo(self.spack_config.env.name, self.env_path,
                           self.spack_config.env.git_path, self.spack_config.env.git_branch,
                           logger=self.logger)
        elif self.env_path and self.env_path.exists():
            self.logger.info('Setup spack environment: the spack environment is already created and fetched.')
            return
        else:
            os.makedirs(self.env_path, exist_ok=True)
            if self.env_path:
                run_command("bash", "-c",
                            f'{self.spack_setup_script} && spack env create -d {self.env_path}',
                            check=True, logger=self.logger,
                            info_msg=f"Created {self.spack_config.env.name} spack environment",
                            exception_msg=f"Failed to create {self.spack_config.env.name} spack environment",
                            exception=BashCommandException)
                self.logger.info('Setup spack environment: created spack environment')
            else:
                raise MissingAttributeException(f'Missing env_path attribute class {__name__}')

    def merge_envs(self, env_path: Path, site_config_path: Path):
        """
        Merges current environment with the specified environment from site-config
        """
        site_config_dest = self.env_path / 'site-config'
        if site_config_dest.exists():
            shutil.rmtree(site_config_dest)
            os.makedirs(site_config_dest)

        self.logger.debug(f'site-config path: {site_config_path.parent}')
        shutil.copytree(site_config_path.parent.parent, site_config_dest)

        script_dir = Path(__file__).resolve().parent.parent
        y_merge_path = script_dir / 'utils' / 'ymerge.py'
        self.logger.debug(f'Original env path: {env_path}')
        self.logger.debug(f'Additional env path: {site_config_path}')
        # update environment site-configs
        merged_envs = run_command(
            "bash", "-c",
            f'{self.spack_setup_script} && spack-python {y_merge_path} {site_config_path} {env_path}',
            info_msg='Merging top-level and site-specific spack.yaml files.',
            exception_msg='Failed to merge top-level and site-specific spack.yaml files.',
            capture_output=True,
            exception=SpackMergeEnvsException,
            text=True,
            check=True
        )

        tmp_spack_yaml = Path("/tmp/spack.yaml").resolve()
        try:
            with open(tmp_spack_yaml, "w") as f:
                f.write(merged_envs.stdout)
        except Exception as e:
            raise SpackMergeEnvsException(f"Error writing {tmp_spack_yaml}: {e}\n")
        if os.path.exists(self.env_path / 'spack.yaml'):
            os.remove(self.env_path / 'spack.yaml')
        shutil.copy(tmp_spack_yaml, self.env_path)

    def spack_repo_exists(self, repo_name: str) -> bool | None:
        """Check if the given Spack repository exists.
        Returns:
            True if spack repository exists, False otherwise.
        """
        if self.spack_config.env is None:
            result = run_command("bash", "-c",
                                 f'{self.spack_setup_script} && spack repo list',
                                 check=True,
                                 capture_output=True, text=True, logger=self.logger,
                                 info_msg=f'Checking if {repo_name} exists')
            if result is None:
                return False
        else:
            if self.spack_env_exists():
                result = run_command("bash", "-c",
                                     f'{self.spack_command_on_env} && spack repo list',
                                     check=True,
                                     capture_output=True, text=True, logger=self.logger,
                                     info_msg=f'Checking if repository {repo_name} was added').stdout
            else:
                self.logger.debug('No spack environment defined')
                raise NoSpackEnvironmentException('No spack environment defined')
            if result is None:
                return False
        return any(line.strip().endswith(repo_name) for line in result.splitlines())

    def spack_env_exists(self):
        """Checks if a spack environments exists.
        Returns:
            True if spack environments exists, False otherwise.
        """
        result = run_command("bash", "-c",
                             self.spack_command_on_env,
                             check=True,
                             capture_output=True, text=True, logger=self.logger,
                             info_msg=f'Checking if environment {self.spack_config.env.name} exists')
        return result is not None

    def add_spack_repo(self, repo_path: Path, repo_name: str):
        """Add the Spack repository if it does not exist."""
        run_command("bash", "-c",
                    f'{self.spack_command_on_env} && spack repo add {repo_path}/{repo_name}',
                    check=True, logger=self.logger,
                    info_msg=f"Added {repo_name} to spack environment {self.spack_config.env.name}",
                    exception_msg=f"Failed to add {repo_name} to spack environment {self.spack_config.env.name}",
                    exception=SpackRepoException)

    def get_compiler_version(self):
        """Returns the compiler version
        Raises:
            NoSpackEnvironmentException: If the spack environment is not set up.
        """
        result = run_command("bash", "-c",
                             f'{self.spack_command_on_env} && spack compiler list',
                             check=True, logger=self.logger,
                             capture_output=True, text=True,
                             info_msg=f"Checking spack environment compiler version for {self.spack_config.env.name}",
                             exception_msg=f"Failed to checking spack environment compiler version for {self.spack_config.env.name}",
                             exception=BashCommandException)

        if result.stdout is None:
            self.logger.debug(f'No gcc found for {self.spack_config.env.name}')
            return None

        # Find the first occurrence of a GCC compiler using regex
        match = re.search(r"gcc@([\d\.]+)", result.stdout)
        gcc_version = match.group(1)
        self.logger.debug(f'Found gcc for {self.spack_config.env.name}: {gcc_version}')
        return gcc_version

    def config(self, config_type: SpackConfigCommand, config_parameter):
        run_command("bash", "-c",
                    f'{self.spack_command_on_env} && spack config {config_type.value} \"{config_parameter}\"',
                    check=True,
                    logger=self.logger,
                    info_msg='Spack config command',
                    exception_msg='Spack config command failed',
                    exception=SpackConfigException)

    def add_package(self, package: str):
        run_command("bash", "-c",
                    f'{self.spack_command_on_env} && spack add {package}"',
                    check=True,
                    logger=self.logger,
                    info_msg=f'Spack add {package} command',
                    exception_msg=f'Spack add  command {package} failed',
                    exception=SpackAddException)

    def remove_package(self, package: str, force: bool = False):
        force = '--force' if force else ''
        run_command("bash", "-c",
                    f'{self.spack_command_on_env} && spack uninstall {force}  {package}"',
                    check=True,
                    logger=self.logger,
                    info_msg=f'Spack remove {package} command',
                    exception_msg=f'Spack remove  command {package} failed',
                    exception=SpackRemoveException)

    def find_packages(self):
        """Returns a dictionary of installed Spack packages in the current environment.
        Each key is the name of a Spack package, and the corresponding value is a list of
        installed versions for that package.
        Raises:
            NoSpackEnvironmentException: If the spack environment is not set up.
        """
        packages = run_command("bash", "-c",
                               f'{self.spack_command_on_env} && spack find -c',
                               check=True,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE,
                               text=True,
                               logger=self.logger,
                               info_msg=f'Listing installed packages.',
                               exception_msg=f'Failed to list installed packages',
                               exception=SpackFindException).stdout
        dict_packages = {}
        for package in packages.strip().splitlines():
            line = package.strip()
            if line.startswith(('[+]', '[e]')):
                parts = line.replace('@', ' ').split()
                if len(parts) == 3:
                    _, name, version = parts
                    dict_packages.setdefault(name, []).append(version)
        return dict_packages

    def find_package(self, package, long: bool, variants: bool, paths: bool):
        """
            Searches for a specific packages
        """
        if package is None:
            raise SpackFindException("No package was passed as argument.")

        long = '--long' if long else ''
        variants = '--variants' if variants else ''
        paths = '--paths' if paths else ''
        package = run_command("bash", "-c",
                              f'{self.spack_command_on_env} && spack find {long} {variants} {paths} {package}',
                              check=True,
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE,
                              text=True,
                              logger=self.logger,
                              info_msg=f'Finding package {package}',
                              exception_msg=f'Failed find package {package}',
                              exception=SpackFindException).stdout
        return package

    def fetch(self, dependencies=False, missing=False):
        """Spack fetch functionality.
        Raises:
            SpackFetchException: If the spack fails to fetch.
        """
        dependencies = '--dependencies' if dependencies else ''
        missing = '--missing' if missing else ''
        result = run_command("bash", "-c",
                             f'{self.spack_command_on_env} && spack fetch {dependencies} {missing}',
                             check=True, logger=self.logger,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE,
                             capture_output=True,
                             text=True,
                             info_msg=f"Checking spack environment compiler version for {self.spack_config.env.name}",
                             exception_msg=f"Failed to checking spack environment compiler version for {self.spack_config.env.name}",
                             exception=SpackFetchException)
        return result

    def spec_pacakge(self, package_name: str, full_format=False):
        """Spec step for a spack environment
            Raises:
                SpackSpecException: If the spack spec command fails.
        """
        try:
            spec_output = run_command("bash", "-c",
                                      f'{self.spack_command_on_env} && spack spec {package_name}',
                                      check=True,
                                      stdout=subprocess.PIPE,
                                      stderr=subprocess.PIPE,
                                      text=True,
                                      logger=self.logger,
                                      info_msg=f'Spack spec {package_name}.',
                                      exception_msg=f'Failed to spack spec {package_name}.',
                                      exception=SpackSpecException).stdout
            if full_format:
                return spec_output
            else:
                pattern = r'^\s*-\s*([\w.-]+@[\d.]+)'
                match = re.search(pattern, spec_output)
                if match:
                    return match.group(1)
                return None
        except SpackSpecException:
            return None

    def check_installed_spack_packages(self, env_path: Path):
        """
            Checks if all spack packages from the environment (spack.yaml) were installed successfully.
         """
        to_install = extract_spack_packages(env_path / 'spack.yaml')
        installed = self.find_packages()
        installed = list(installed.keys())
        for package in to_install:
            if package not in installed:
                return False
        return True

    def create_load_env_script(self):
        """
            Creates load script that when sourced activates and loads the installed spack environment, using views
        """
        with open(self.env_path / 'load_env.sh', "w") as load_env:
            run_command("bash", "-c",
                        f'{self.spack_setup_script} && spack env activate --sh {self.env_path}',
                        check=True,
                        stdout=load_env,
                        stderr=subprocess.PIPE,
                        exception=CreateLoadEnvException,
                        text=True,
                        )

    def spack_clean(self, bootstrap: bool = False):
        bootstrap = '--bootstrap' if bootstrap else ''
        run_command("bash", "-c",
                    f'{self.spack_setup_script} && spack clean -a {bootstrap}',
                    check=True,
                    logger=self.logger,
                    info_msg=f'Spack clean command',
                    exception_msg=f'Spack failed to clean command failed',
                    exception=SpackRemoveException)



from dedal.configuration.SpackConfig import SpackConfig
from dedal.model.SpackDescriptor import SpackDescriptor
from dedal.spack_factory.SpackOperation import SpackOperation
from dedal.spack_factory.SpackOperationCreator import SpackOperationCreator


def test_serialize_and_deserialize_spack_operation(tmp_path):
    install_dir = tmp_path
    env = SpackDescriptor('test_spack_env', install_dir)
    config = SpackConfig(env=env, install_dir=install_dir, serialize_name='data_test.pkl')
    spack_operation = SpackOperationCreator.get_spack_operator(config)
    spack_operation.serialize()
    loaded_spack_operation: SpackOperation = SpackOperation.deserialize(install_dir, "data_test.pkl")
    assert isinstance(loaded_spack_operation, SpackOperation)
    assert loaded_spack_operation.spack_config.env.name == spack_operation.spack_config.env.name



# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

ebrains_spack_builds_git = 'https://gitlab.ebrains.eu/ri/tech-hub/platform/esd/ebrains-spack-builds.git'
SPACK_VERSION = "0.23.1"
SPACK_ENV_ACCESS_TOKEN = os.getenv("SPACK_ENV_ACCESS_TOKEN")
test_spack_env_git = f'https://oauth2:{SPACK_ENV_ACCESS_TOKEN}@gitlab.ebrains.eu/ri/projects-and-initiatives/virtualbraintwin/tools/test-spack-env.git'



# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

class SpackException(Exception):

    def __init__(self, message):
        super().__init__(message)
        self.message = str(message)

    def __str__(self):
        return self.message


class BashCommandException(SpackException):
    """
    To be thrown when a bash command has failed
    """


class NoSpackEnvironmentException(BashCommandException):
    """
    To be thrown when an operation on a spack environment is executed without the environment being activated or existent
    """


class SpackConcertizeException(BashCommandException):
    """
    To be thrown when the spack concretization step fails
    """


class SpackInstallPackagesException(BashCommandException):
    """
    To be thrown when the spack fails to install spack packages
    """


class SpackMirrorException(BashCommandException):
    """
    To be thrown when the spack add mirror command fails
    """


class SpackGpgException(BashCommandException):
    """
    To be thrown when the spack fails to create gpg keys
    """


class SpackRepoException(BashCommandException):
    """
    To be thrown when the spack fails to add a repo
    """


class SpackReindexException(BashCommandException):
    """
    To be thrown when the spack reindex step fails
    """


class SpackSpecException(BashCommandException):
    """
    To be thrown when the spack spec for a package fails
    """


class SpackConfigException(BashCommandException):
    """
    To be thrown when the spack config command fails
    """


class SpackFindException(BashCommandException):
    """
    To be thrown when the spack find command fails
    """


class MissingAttributeException(BashCommandException):
    """
    To be thrown when a missing attribute for a class is missing
    """


class SpackFetchException(BashCommandException):
    """
    To be thrown when a fetching fails
    """


class SpackCreateBuildCacheException(BashCommandException):
    """
    To be thrown when a creating a buildcache for a package fails
    """


class SpackMergeEnvsException(BashCommandException):
    """
    To be thrown when merging spack environments fails
    """


class CreateLoadEnvException(BashCommandException):
    """
    To be thrown when creating the load env bash script fails
    """


class SpackAddException(BashCommandException):
    """
    To be thrown when spack add <package> fails
    """


class SpackRemoveException(BashCommandException):
    """
    To be thrown when spack remove <package> fails
    """



# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pathlib import Path

from dedal.configuration.SpackConfig import SpackConfig
from dedal.model.SpackDescriptor import SpackDescriptor
from dedal.spack_factory.SpackOperationCreator import SpackOperationCreator
from dedal.spack_factory.SpackOperationUseCache import SpackOperationUseCache
from dedal.tests.integration_tests.helper import get_test_path
from dedal.utils.utils import file_exists_and_not_empty, count_files_in_folder
from dedal.utils.variables import test_spack_env_git, ebrains_spack_builds_git


def setup(path):
    install_dir = path
    env_name = 'test-spack-env'
    env_path = install_dir / env_name
    env = SpackDescriptor(env_name, install_dir, test_spack_env_git)
    repo = SpackDescriptor('ebrains-spack-builds', install_dir, ebrains_spack_builds_git)
    concretization_dir = install_dir / 'concretize'
    buildcache_dir = install_dir / 'buildcache'
    spack_config = SpackConfig(env, install_dir=install_dir, concretization_dir=concretization_dir,
                               buildcache_dir=buildcache_dir)
    spack_config.add_repo(repo)
    spack_operation = SpackOperationCreator.get_spack_operator(spack_config, use_cache=True)
    return spack_operation, concretization_dir, buildcache_dir, env_path


def test_spack_from_cache_setup():
    install_dir = get_test_path()
    spack_operation, concretization_dir, buildcache_dir, _ = setup(install_dir)
    assert isinstance(spack_operation, SpackOperationUseCache)
    spack_operation.install_spack(bashrc_path=str(install_dir / Path('.bashrc')))
    spack_operation.setup_spack_env()
    num_tags = len(spack_operation.build_cache.list_tags())
    concretization_download_file_path = concretization_dir / 'spack.lock'
    assert file_exists_and_not_empty(concretization_download_file_path) == True
    assert count_files_in_folder(buildcache_dir) == num_tags + 2
    assert 'local_cache' in spack_operation.mirror_list()


def test_spack_from_cache_concretize():
    install_dir = get_test_path()
    spack_operation, concretization_dir, buildcache_dir, _ = setup(install_dir)
    assert isinstance(spack_operation, SpackOperationUseCache)
    spack_operation.install_spack(bashrc_path=str(install_dir / Path('.bashrc')))
    spack_operation.setup_spack_env()
    num_tags = len(spack_operation.build_cache.list_tags())
    concretization_download_file_path = concretization_dir / 'spack.lock'
    assert file_exists_and_not_empty(concretization_download_file_path) == True
    assert count_files_in_folder(buildcache_dir) == num_tags + 2
    assert 'local_cache' in spack_operation.mirror_list()
    assert spack_operation.concretize_spack_env() == False
    concretization_file_path = spack_operation.env_path / 'spack.lock'
    assert file_exists_and_not_empty(concretization_file_path) == True


def test_spack_from_cache_install_1():
    install_dir = get_test_path()
    spack_operation, concretization_dir, buildcache_dir, env_path = setup(install_dir)
    assert isinstance(spack_operation, SpackOperationUseCache)
    spack_operation.install_spack(bashrc_path=str(install_dir / Path('.bashrc')))
    spack_operation.setup_spack_env()
    num_tags = len(spack_operation.build_cache.list_tags())
    concretization_download_file_path = concretization_dir / 'spack.lock'
    assert file_exists_and_not_empty(concretization_download_file_path) == True
    assert count_files_in_folder(buildcache_dir) == num_tags + 2
    assert 'local_cache' in spack_operation.mirror_list()
    assert spack_operation.concretize_spack_env() == False
    concretization_file_path = spack_operation.env_path / 'spack.lock'
    assert file_exists_and_not_empty(concretization_file_path) == True
    install_result = spack_operation.install_packages(jobs=2, debug=False)
    assert install_result.returncode == 0
    assert spack_operation.check_installed_spack_packages(env_path) == True


def test_spack_from_cache_install_2():
    install_dir = get_test_path()
    spack_operation, concretization_dir, buildcache_dir, env_path = setup(install_dir)
    assert isinstance(spack_operation, SpackOperationUseCache)
    spack_operation.install_spack(bashrc_path=str(install_dir / Path('.bashrc')))
    spack_operation.setup_spack_env()
    num_tags = len(spack_operation.build_cache.list_tags())
    concretization_download_file_path = concretization_dir / 'spack.lock'
    assert file_exists_and_not_empty(concretization_download_file_path) == True
    assert count_files_in_folder(buildcache_dir) == num_tags + 2
    assert 'local_cache' in spack_operation.mirror_list()
    assert spack_operation.concretize_spack_env() == False
    concretization_file_path = spack_operation.env_path / 'spack.lock'
    assert file_exists_and_not_empty(concretization_file_path) == True
    install_result = spack_operation.install_packages(jobs=2, debug=False, test='root')
    assert install_result.returncode == 0
    assert spack_operation.check_installed_spack_packages(env_path) == True



# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pathlib import Path
import pytest
from dedal.configuration.SpackConfig import SpackConfig
from dedal.error_handling.exceptions import BashCommandException, NoSpackEnvironmentException
from dedal.spack_factory.SpackOperationCreator import SpackOperationCreator
from dedal.model.SpackDescriptor import SpackDescriptor
from dedal.tests.testing_variables import test_spack_env_git, ebrains_spack_builds_git
from dedal.utils.utils import file_exists_and_not_empty, git_clone_repo
from dedal.spack_factory.SpackOperation import SpackOperation


def test_spack_repo_exists_1(tmp_path):
    install_dir = tmp_path
    env = SpackDescriptor('ebrains-spack-builds', install_dir)
    config = SpackConfig(env=env, install_dir=install_dir)
    spack_operation = SpackOperationCreator.get_spack_operator(config)
    spack_operation.install_spack(bashrc_path=str(tmp_path / Path('.bashrc')))
    with pytest.raises(NoSpackEnvironmentException):
        spack_operation.spack_repo_exists(env.name)


def test_spack_repo_exists_2(tmp_path):
    install_dir = tmp_path
    env = SpackDescriptor('ebrains-spack-builds', install_dir)
    config = SpackConfig(env=env, install_dir=install_dir)
    spack_operation = SpackOperationCreator.get_spack_operator(config)
    assert isinstance(spack_operation, SpackOperation)
    spack_operation.install_spack(bashrc_path=str(tmp_path / Path('.bashrc')))
    spack_operation.setup_spack_env()
    assert spack_operation.spack_repo_exists(env.name) == False


def test_spack_from_scratch_setup_1(tmp_path):
    install_dir = tmp_path
    env = SpackDescriptor('ebrains-spack-builds', install_dir, ebrains_spack_builds_git)
    config = SpackConfig(env=env, system_name='ebrainslab', install_dir=install_dir)
    spack_operation = SpackOperationCreator.get_spack_operator(config)
    assert isinstance(spack_operation, SpackOperation)
    spack_operation.install_spack(bashrc_path=str(tmp_path / Path('.bashrc')))
    spack_operation.setup_spack_env()
    assert spack_operation.spack_repo_exists(env.name) == False


def test_spack_reindex(tmp_path):
    install_dir = tmp_path
    config = SpackConfig(install_dir=install_dir)
    spack_operation = SpackOperationCreator.get_spack_operator(config)
    assert isinstance(spack_operation, SpackOperation)
    spack_operation.install_spack(bashrc_path=str(tmp_path / Path('.bashrc')))
    spack_operation.reindex()


@pytest.mark.skip(reason="It does not work on bare metal operating systems")
def test_spack_spec(tmp_path):
    install_dir = tmp_path
    config = SpackConfig(install_dir=install_dir)
    spack_operation = SpackOperationCreator.get_spack_operator(config)
    assert isinstance(spack_operation, SpackOperation)
    spack_operation.install_spack(bashrc_path=str(tmp_path / Path('.bashrc')))
    assert spack_operation.spec_pacakge('aida') == 'aida@3.2.1'


def test_spack_from_scratch_setup_2(tmp_path):
    install_dir = tmp_path
    env = SpackDescriptor('ebrains-spack-builds', install_dir, ebrains_spack_builds_git)
    repo = env
    config = SpackConfig(env=env, system_name='ebrainslab', install_dir=install_dir)
    config.add_repo(repo)
    config.add_repo(repo)
    spack_operation = SpackOperationCreator.get_spack_operator(config)
    assert isinstance(spack_operation, SpackOperation)
    spack_operation.install_spack(bashrc_path=str(tmp_path / Path('.bashrc')))
    spack_operation.setup_spack_env()
    assert spack_operation.spack_repo_exists(env.name) == True


def  test_spack_from_scratch_setup_3(tmp_path):
    install_dir = tmp_path
    repo_name = 'ebrains-spack-builds'
    git_clone_repo(repo_name, install_dir / repo_name, ebrains_spack_builds_git)
    env = SpackDescriptor(repo_name, install_dir)
    repo = env
    config = SpackConfig(env=env, system_name='ebrainslab', install_dir=install_dir)
    config.add_repo(repo)
    spack_operation = SpackOperationCreator.get_spack_operator(config)
    assert isinstance(spack_operation, SpackOperation)
    spack_operation.install_spack(bashrc_path=str(tmp_path / Path('.bashrc')))
    spack_operation.setup_spack_env()
    assert spack_operation.spack_repo_exists(env.name) == True


def test_spack_from_scratch_setup_4(tmp_path):
    install_dir = tmp_path
    env = SpackDescriptor('new_env1', install_dir)
    repo = env
    config = SpackConfig(env=env, system_name='ebrainslab', install_dir=install_dir)
    config.add_repo(repo)
    config.add_repo(repo)
    spack_operation = SpackOperationCreator.get_spack_operator(config)
    assert isinstance(spack_operation, SpackOperation)
    spack_operation.install_spack(bashrc_path=str(tmp_path / Path('.bashrc')))
    with pytest.raises(BashCommandException):
        spack_operation.setup_spack_env()


def test_spack_from_scratch_setup_5(tmp_path):
    install_dir = tmp_path
    env = SpackDescriptor('new_env2', install_dir)
    config = SpackConfig(env=env, install_dir=install_dir)
    spack_operation = SpackOperationCreator.get_spack_operator(config)
    assert isinstance(spack_operation, SpackOperation)
    spack_operation.install_spack(bashrc_path=str(tmp_path / Path('.bashrc')))
    spack_operation.setup_spack_env()
    assert spack_operation.spack_env_exists() == True


def test_spack_not_a_valid_repo():
    env = SpackDescriptor('ebrains-spack-builds', Path(), None)
    repo = env
    config = SpackConfig(env=env, system_name='ebrainslab')
    config.add_repo(repo)
    spack_operation = SpackOperationCreator.get_spack_operator(config)
    assert isinstance(spack_operation, SpackOperation)
    with pytest.raises(BashCommandException):
        spack_operation.add_spack_repo(repo.path, repo.name)


@pytest.mark.skip(
    reason="Skipping the concretization step because it may freeze when numerous Spack packages are added to the environment.")
def test_spack_from_scratch_concretize_1(tmp_path):
    install_dir = tmp_path
    env = SpackDescriptor('ebrains-spack-builds', install_dir, ebrains_spack_builds_git)
    repo = env
    config = SpackConfig(env=env, system_name='ebrainslab', install_dir=install_dir)
    config.add_repo(repo)
    config.add_repo(repo)
    spack_operation = SpackOperationCreator.get_spack_operator(config)
    assert isinstance(spack_operation, SpackOperation)
    spack_operation.install_spack(bashrc_path=str(tmp_path / Path('.bashrc')))
    spack_operation.install_spack(bashrc_path=str(tmp_path / Path('.bashrc')))
    spack_operation.setup_spack_env()
    spack_operation.concretize_spack_env(force=True)
    concretization_file_path = spack_operation.env_path / 'spack.lock'
    assert file_exists_and_not_empty(concretization_file_path) == True


@pytest.mark.skip(
    reason="Skipping the concretization step because it may freeze when numerous Spack packages are added to the environment.")
def test_spack_from_scratch_concretize_2(tmp_path):
    install_dir = tmp_path
    env = SpackDescriptor('ebrains-spack-builds', install_dir, ebrains_spack_builds_git)
    repo = env
    config = SpackConfig(env=env, system_name='ebrainslab', install_dir=install_dir)
    config.add_repo(repo)
    config.add_repo(repo)
    spack_operation = SpackOperationCreator.get_spack_operator(config)
    assert isinstance(spack_operation, SpackOperation)
    spack_operation.install_spack(bashrc_path=str(tmp_path / Path('.bashrc')))
    spack_operation.setup_spack_env()
    spack_operation.concretize_spack_env(force=False)
    concretization_file_path = spack_operation.env_path / 'spack.lock'
    assert file_exists_and_not_empty(concretization_file_path) == True


def test_spack_from_scratch_concretize_3(tmp_path):
    install_dir = tmp_path
    env = SpackDescriptor('ebrains-spack-builds', install_dir, ebrains_spack_builds_git)
    repo = env
    config = SpackConfig(env=env, system_name='ebrainslab', install_dir=install_dir)
    config.add_repo(repo)
    config.add_repo(repo)
    spack_operation = SpackOperationCreator.get_spack_operator(config)
    assert isinstance(spack_operation, SpackOperation)
    spack_operation.install_spack(bashrc_path=str(tmp_path / Path('.bashrc')))
    spack_operation.setup_spack_env()
    concretization_file_path = spack_operation.env_path / 'spack.lock'
    assert file_exists_and_not_empty(concretization_file_path) == False


def test_spack_from_scratch_concretize_4(tmp_path):
    install_dir = tmp_path
    env = SpackDescriptor('test-spack-env', install_dir, test_spack_env_git)
    config = SpackConfig(env=env, install_dir=install_dir)
    spack_operation = SpackOperationCreator.get_spack_operator(config)
    assert isinstance(spack_operation, SpackOperation)
    spack_operation.install_spack(bashrc_path=str(tmp_path / Path('.bashrc')))
    spack_operation.setup_spack_env()
    spack_operation.concretize_spack_env(force=False)
    concretization_file_path = spack_operation.env_path / 'spack.lock'
    assert file_exists_and_not_empty(concretization_file_path) == True


def test_spack_from_scratch_concretize_5(tmp_path):
    install_dir = tmp_path
    env = SpackDescriptor('test-spack-env', install_dir, test_spack_env_git)
    config = SpackConfig(env=env, install_dir=install_dir)
    spack_operation = SpackOperationCreator.get_spack_operator(config)
    assert isinstance(spack_operation, SpackOperation)
    spack_operation.install_spack(bashrc_path=str(tmp_path / Path('.bashrc')))
    spack_operation.setup_spack_env()
    spack_operation.concretize_spack_env(force=True)
    concretization_file_path = spack_operation.env_path / 'spack.lock'
    assert file_exists_and_not_empty(concretization_file_path) == True


def test_spack_from_scratch_concretize_6(tmp_path):
    install_dir = tmp_path
    env = SpackDescriptor('test-spack-env', install_dir, test_spack_env_git)
    repo = SpackDescriptor('ebrains-spack-builds', install_dir, ebrains_spack_builds_git)
    config = SpackConfig(env=env, install_dir=install_dir)
    config.add_repo(repo)
    spack_operation = SpackOperationCreator.get_spack_operator(config)
    assert isinstance(spack_operation, SpackOperation)
    spack_operation.install_spack(bashrc_path=str(tmp_path / Path('.bashrc')))
    spack_operation.setup_spack_env()
    spack_operation.concretize_spack_env(force=False)
    concretization_file_path = spack_operation.env_path / 'spack.lock'
    assert file_exists_and_not_empty(concretization_file_path) == True


def test_spack_from_scratch_concretize_7(tmp_path):
    install_dir = tmp_path
    env = SpackDescriptor('test-spack-env', install_dir, test_spack_env_git)
    repo = SpackDescriptor('ebrains-spack-builds', install_dir, ebrains_spack_builds_git)
    config = SpackConfig(env=env, install_dir=install_dir)
    config.add_repo(repo)
    spack_operation = SpackOperationCreator.get_spack_operator(config)
    assert isinstance(spack_operation, SpackOperation)
    spack_operation.install_spack(bashrc_path=str(tmp_path / Path('.bashrc')))
    spack_operation.setup_spack_env()
    spack_operation.concretize_spack_env(force=True)
    concretization_file_path = spack_operation.env_path / 'spack.lock'
    assert file_exists_and_not_empty(concretization_file_path) == True


def test_spack_from_scratch_concretize_8(tmp_path):
    install_dir = tmp_path
    env = SpackDescriptor('test-spack-env', install_dir, test_spack_env_git)
    repo = SpackDescriptor('ebrains-spack-builds', install_dir, ebrains_spack_builds_git)
    config = SpackConfig(env=env, install_dir=install_dir)
    config.add_repo(repo)
    spack_operation = SpackOperationCreator.get_spack_operator(config)
    assert isinstance(spack_operation, SpackOperation)
    spack_operation.install_spack(bashrc_path=str(tmp_path / Path('.bashrc')))
    spack_operation.setup_spack_env()
    spack_operation.concretize_spack_env(force=True, test='root')
    concretization_file_path = spack_operation.env_path / 'spack.lock'
    assert file_exists_and_not_empty(concretization_file_path) == True


def test_spack_from_scratch_install(tmp_path):
    install_dir = tmp_path
    env_name = 'test-spack-env'
    env_path = install_dir / env_name
    env = SpackDescriptor(env_name, install_dir, test_spack_env_git)
    repo = SpackDescriptor('ebrains-spack-builds', install_dir, ebrains_spack_builds_git)
    config = SpackConfig(env=env, install_dir=install_dir)
    config.add_repo(repo)
    spack_operation = SpackOperationCreator.get_spack_operator(config)
    assert isinstance(spack_operation, SpackOperation)
    spack_operation.install_spack(bashrc_path=str(tmp_path / Path('.bashrc')))
    spack_operation.setup_spack_env()
    spack_operation.concretize_spack_env(force=True)
    concretization_file_path = spack_operation.env_path / 'spack.lock'
    assert file_exists_and_not_empty(concretization_file_path) == True
    install_result = spack_operation.install_packages(jobs=2, signed=False, fresh=True, debug=False)
    assert install_result.returncode == 0
    assert spack_operation.check_installed_spack_packages(env_path) == True


def test_spack_from_scratch_install_2(tmp_path):
    install_dir = tmp_path
    env_name = 'test-spack-env'
    env_path = install_dir / env_name
    env = SpackDescriptor(env_name, install_dir, test_spack_env_git)
    repo = SpackDescriptor('ebrains-spack-builds', install_dir, ebrains_spack_builds_git)
    config = SpackConfig(env=env, install_dir=install_dir)
    config.add_repo(repo)
    spack_operation = SpackOperationCreator.get_spack_operator(config)
    assert isinstance(spack_operation, SpackOperation)
    spack_operation.install_spack(bashrc_path=str(tmp_path / Path('.bashrc')))
    spack_operation.setup_spack_env()
    spack_operation.concretize_spack_env(force=True, test='root')
    concretization_file_path = spack_operation.env_path / 'spack.lock'
    assert file_exists_and_not_empty(concretization_file_path) == True
    install_result = spack_operation.install_packages(jobs=2, signed=False, fresh=True, debug=False, test='root')
    assert install_result.returncode == 0
    assert spack_operation.check_installed_spack_packages(env_path) == True


def test_spack_mirror_env(tmp_path):
    install_dir = tmp_path
    env = SpackDescriptor('test-spack-env', install_dir, test_spack_env_git)
    repo = SpackDescriptor('ebrains-spack-builds', install_dir, ebrains_spack_builds_git)
    spack_config = SpackConfig(env, install_dir=install_dir)
    spack_config.add_repo(repo)
    spack_operation = SpackOperationCreator.get_spack_operator(spack_config)
    assert isinstance(spack_operation, SpackOperation)
    spack_operation.install_spack(bashrc_path=str(tmp_path / Path('.bashrc')))
    spack_operation.setup_spack_env()
    mirror_dir = tmp_path / Path('./mirror_dir')
    mirror_name = 'mirror_tests'
    spack_operation.add_mirror(mirror_name=mirror_name, mirror_path=mirror_dir)
    assert mirror_name in spack_operation.mirror_list()
    spack_operation.remove_mirror(mirror_name=mirror_name)
    assert mirror_name not in spack_operation.mirror_list()


def test_spack_mirror_global(tmp_path):
    install_dir = tmp_path
    spack_config = SpackConfig(install_dir=install_dir)
    spack_operation = SpackOperationCreator.get_spack_operator(spack_config)
    assert isinstance(spack_operation, SpackOperation)
    spack_operation.install_spack(bashrc_path=str(tmp_path / Path('.bashrc')))
    spack_operation.setup_spack_env()
    mirror_dir = tmp_path / Path('./mirror_dir')
    mirror_name = 'mirror_test'
    spack_operation.add_mirror(mirror_name=mirror_name, mirror_path=mirror_dir)
    assert mirror_name in spack_operation.mirror_list()
    spack_operation.remove_mirror(mirror_name=mirror_name)
    assert mirror_name not in spack_operation.mirror_list()



# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import pathlib
from typing import List
import spack
import ruamel.yaml as yaml
import llnl.util.filesystem as fsys
from collections.abc import Iterable

# migration of /cli/specfile_storage_path_source.py
class SpecfileSourceStoragePathExtractor:
    def __init__(self, path_specfile: pathlib.Path, include_installed: bool = False):
        self.path_specfile = path_specfile
        self.include_installed = include_installed

    def extract_paths(self) -> List[str]:
        with open(self.path_specfile, "r") as fd:
            file_content = fd.read()
            data = list(yaml.safe_load_all(file_content))

        to_be_fetched = set()
        for rspec in data:
            s = spack.spec.Spec.from_dict(rspec)
            if not isinstance(s, Iterable):
                s = [s]

            maybe_to_be_fetched = spack.traverse.traverse_nodes(s, key=spack.traverse.by_dag_hash)

            for ss in maybe_to_be_fetched:
                if (not self.include_installed) and ss.installed:
                    continue

                pkg = ss.package

                if not pkg.fetcher.cachable:
                    continue

                format_string = "{name}-{version}"
                pretty_name = pkg.spec.format_path(format_string)
                cosmetic_path = os.path.join(pkg.name, pretty_name)
                to_be_fetched.add(str(spack.mirror.mirror_archive_paths(pkg.fetcher, cosmetic_path).storage_path))
                for resource in pkg._get_needed_resources():
                    pretty_resource_name = fsys.polite_filename(f"{resource.name}-{pkg.version}")
                    to_be_fetched.add(str(spack.mirror.mirror_archive_paths(resource.fetcher, pretty_resource_name).storage_path))
                for patch in ss.patches:
                    if isinstance(patch, spack.patch.UrlPatch):
                        to_be_fetched.add(str(spack.mirror.mirror_archive_paths(patch.stage.fetcher, patch.stage.name).storage_path))

        return list(to_be_fetched)


# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
from pathlib import Path
import sys
from typing import Dict, List
from dedal.build_cache.BuildCacheManagerOciDefault import BuildCacheManagerOciDefault

default_remote_cache = (("HARBOR_HOST" in os.environ) and ("HARBOR_PROJECT" in os.environ)) and (
    "{}/{}/build_cache".format(os.environ["HARBOR_HOST"], os.environ["HARBOR_PROJECT"])) or ""
default_remote_cache_username = "HARBOR_USERNAME" in os.environ and Path(os.environ["HARBOR_USERNAME"]) or ""
default_remote_cache_password = "HARBOR_PASSWORD" in os.environ and Path(os.environ["HARBOR_PASSWORD"]) or None
default_local_cache = "YASHCHIKI_CACHE_BUILD" in os.environ and Path(
    os.environ["YASHCHIKI_CACHE_BUILD"]) or os.path.expanduser("~/.yashchiki/cache/")


# migration of /cli/update_cached_buildresults.py
class CachedBuildResultUploader:
    def __init__(self,
                 path_missing: Path,
                 remote_cache: str=default_remote_cache,
                 remote_cache_username: str=default_remote_cache_username,
                 remote_cache_password: str=default_remote_cache_password,
                 remote_cache_type='oci',
                 local_cache: Path=default_local_cache):

        self.path_missing = path_missing
        self.remote_cache = remote_cache
        self.remote_cache_username = remote_cache_username
        self.remote_cache_password = remote_cache_password
        self.local_cache = local_cache
        self.remote_cache_type=remote_cache_type
        # Extract host and project from the remote_cache URL
        try:
            parts = remote_cache.split('/')
            self.registry_host = parts[0]
            self.registry_project = parts[1]
        except Exception:
            raise ValueError(f"Invalid remote_cache format: {remote_cache}")

        self.oci = BuildCacheManagerOciDefault(
            registry_host=self.registry_host,
            registry_project=self.registry_project,
            registry_username=self.remote_cache_username,
            registry_password=self.remote_cache_password
        )

        if self.oci is None:
            raise RuntimeError("Failed to initialize BuildCacheManagerOci")

    def upload_missing_results(self):
        if not self.path_missing.exists():
            print(f"File with missing cached build information is not available: {self.path_missing}")
            sys.exit(0)

        packages: Dict[str, List[str]] = {}
        with open(self.path_missing, "r") as fd:
            for line in fd:
                elems = line.strip().split()
                if elems:
                    packages[elems[0]] = elems[1:]

        for package_dag_hash, paths in packages.items():
            for path in paths:
                basename = os.path.basename(path)
                full_path = self.local_cache / path

                if ((str(full_path).endswith(".spack") or str(full_path).endswith(
                        ".spec.json")) and not full_path.exists()):
                    print(f"Missing local cache entry for \"{full_path}\"")
                    continue

                if not full_path.exists():
                    # Skip other non-existent files
                    continue

                # Upload file using BuildCacheManagerOci's internal ORAS client
                try:
                    self.oci.client.push(
                        files=[str(full_path)],
                        target=f"{self.remote_cache}:{basename}",
                        manifest_annotations={"path": path},
                        disable_path_validation=True,
                    )
                    print(f"Successfully uploaded {basename} to {self.remote_cache}:{basename}")
                except Exception as e:
                    print(f"Uploading of \"{path}\" to \"{self.remote_cache}:{basename}\" failed: {e}")


import pickle
from pathlib import Path
from typing import Any

from dedal.serialize.SerializeInterface import SerializerInterface


class PickleSerializer(SerializerInterface):
    def __init__(self, file_path: Path, file_name: str) -> None:
        self.file = file_path / file_name

    def serialize(self, obj: Any) -> None:
        """
        Serialize `obj` to disk as `filename` using the highest available protocol.
        """
        with open(self.file, 'wb') as f:
            pickle.dump(obj, f, protocol=pickle.HIGHEST_PROTOCOL)

    def deserialize(self) -> Any:
        """
        Read `filename` from disk and return the original Python object.
        """
        with open(self.file, 'rb') as f:
            return pickle.load(f)



# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

import pytest
from unittest.mock import patch, MagicMock
from click.testing import CliRunner
from dedal.cli.spack_manager_api import show_config, clear_config, install_spack, add_spack_repo, install_packages, \
    setup_spack_env, concretize, set_config
from dedal.enum.SpackViewEnum import SpackViewEnum
from dedal.model.SpackDescriptor import SpackDescriptor


@pytest.fixture
def runner():
    return CliRunner()


@pytest.fixture
def mocked_session_path():
    return '/mocked/tmp/session.json'


@pytest.fixture
def mock_spack_manager():
    mock_spack_manager = MagicMock()
    mock_spack_manager.install_spack = MagicMock()
    mock_spack_manager.add_spack_repo = MagicMock()
    mock_spack_manager.setup_spack_env = MagicMock()
    mock_spack_manager.concretize_spack_env = MagicMock()
    mock_spack_manager.install_packages = MagicMock()
    return mock_spack_manager


@pytest.fixture
def mock_load_config():
    with patch('dedal.cli.spack_manager_api.load_config') as mock_load:
        yield mock_load


@pytest.fixture
def mock_save_config():
    with patch('dedal.cli.spack_manager_api.save_config') as mock_save:
        yield mock_save


@pytest.fixture
def mock_clear_config():
    with patch('dedal.cli.spack_manager_api.clear_config') as mock_clear:
        yield mock_clear


def test_show_config_no_config(runner, mock_load_config):
    mock_load_config.return_value = None
    result = runner.invoke(show_config)
    assert 'No configuration set. Use `set-config` first.' in result.output


def test_show_config_with_config(runner, mock_load_config):
    """Test the show_config command when config is present."""
    mock_load_config.return_value = {"key": "value"}
    result = runner.invoke(show_config)
    assert result.exit_code == 0
    assert '"key": "value"' in result.output


def test_clear_config(runner, mock_clear_config):
    """Test the clear_config command."""
    with patch('os.path.exists', return_value=True), patch('os.remove') as mock_remove:
        result = runner.invoke(clear_config)
        assert 'Configuration cleared!' in result.output
        mock_remove.assert_called_once()


def test_install_spack_no_context_1(runner, mock_spack_manager):
    """Test install_spack with no context, using SpackManager."""
    with patch('dedal.cli.spack_manager_api.SpackManager', return_value=mock_spack_manager):
        result = runner.invoke(install_spack, ['--spack_version', '0.24.0'])
    mock_spack_manager.install_spack.assert_called_once_with('0.24.0', os.path.expanduser("~/.bashrc"))
    assert result.exit_code == 0


def test_install_spack_no_context_2(runner, mock_spack_manager):
    """Test install_spack with no context, using SpackManager and the default value for spack_version."""
    with patch('dedal.cli.spack_manager_api.SpackManager', return_value=mock_spack_manager):
        result = runner.invoke(install_spack)
    mock_spack_manager.install_spack.assert_called_once_with('0.23.0', os.path.expanduser("~/.bashrc"))
    assert result.exit_code == 0


def test_install_spack_with_mocked_context_1(runner, mock_spack_manager):
    """Test install_spack with a mocked context, using ctx.obj as SpackManager."""
    result = runner.invoke(install_spack, ['--spack_version', '0.24.0', '--bashrc_path', '/home/.bahsrc'], obj=mock_spack_manager)
    mock_spack_manager.install_spack.assert_called_once_with('0.24.0', '/home/.bahsrc')
    assert result.exit_code == 0


def test_install_spack_with_mocked_context_2(runner, mock_spack_manager):
    """Test install_spack with a mocked context, using ctx.obj as SpackManager and the default value for spack_version."""
    result = runner.invoke(install_spack, obj=mock_spack_manager)
    mock_spack_manager.install_spack.assert_called_once_with('0.23.0', os.path.expanduser("~/.bashrc"))
    assert result.exit_code == 0


def test_setup_spack_env(runner, mock_spack_manager):
    """Test setup_spack_env with a mocked context, using ctx.obj as SpackManager."""
    result = runner.invoke(setup_spack_env, obj=mock_spack_manager)
    mock_spack_manager.setup_spack_env.assert_called_once_with()
    assert result.exit_code == 0


def test_concretize(runner, mock_spack_manager):
    """Test install_spack with a mocked context, using ctx.obj as SpackManager."""
    result = runner.invoke(concretize, obj=mock_spack_manager)
    mock_spack_manager.concretize_spack_env.assert_called_once_with()
    assert result.exit_code == 0


def test_install_packages_1(runner, mock_spack_manager):
    """Test install_packages with a mocked context, using ctx.obj as SpackManager."""
    result = runner.invoke(install_packages, obj=mock_spack_manager)
    mock_spack_manager.install_packages.assert_called_once_with(jobs=2)
    assert result.exit_code == 0


def test_install_packages(runner, mock_spack_manager):
    """Test install_packages with a mocked context, using ctx.obj as SpackManager."""
    result = runner.invoke(install_packages, ['--jobs', 3], obj=mock_spack_manager)
    mock_spack_manager.install_packages.assert_called_once_with(jobs=3)
    assert result.exit_code == 0


@patch('dedal.cli.spack_manager_api.resolve_path')
@patch('dedal.cli.spack_manager_api.SpackDescriptor')
def test_add_spack_repo(mock_spack_descriptor, mock_resolve_path, mock_load_config, mock_save_config,
                        mocked_session_path, runner):
    """Test adding a spack repository with mocks."""
    expected_config = {'repos': [SpackDescriptor(name='test-repo')]}
    repo_name = 'test-repo'
    path = '/path'
    git_path = 'https://example.com/repo.git'
    mock_resolve_path.return_value = '/resolved/path'
    mock_load_config.return_value = expected_config
    mock_repo_instance = MagicMock()
    mock_spack_descriptor.return_value = mock_repo_instance

    with patch('dedal.cli.spack_manager_api.SESSION_CONFIG_PATH', mocked_session_path):
        result = runner.invoke(add_spack_repo, ['--repo_name', repo_name, '--path', path, '--git_path', git_path])

    assert result.exit_code == 0
    assert 'dedal setup_spack_env must be reran after each repo is added' in result.output
    mock_resolve_path.assert_called_once_with(path)
    mock_spack_descriptor.assert_called_once_with(repo_name, '/resolved/path', git_path)
    assert mock_repo_instance in expected_config['repos']
    mock_save_config.assert_called_once_with(expected_config, mocked_session_path)


def test_set_config(runner, mock_save_config, mocked_session_path):
    """Test set_config."""
    with patch('dedal.cli.spack_manager_api.SESSION_CONFIG_PATH', mocked_session_path):
        result = runner.invoke(set_config, ['--env_name', 'test', '--system_name', 'sys'])

    expected_config = {
        'use_cache': False,
        'env_name': 'test',
        'env_path': None,
        'env_git_path': None,
        'install_dir': None,
        'upstream_instance': None,
        'system_name': 'sys',
        'concretization_dir': None,
        'buildcache_dir': None,
        'gpg_name': None,
        'gpg_mail': None,
        'use_spack_global': False,
        'repos': [],
        'cache_version_concretize': 'v1',
        'cache_version_build': 'v1',
        'view': SpackViewEnum.VIEW,
        'override_cache': True,
    }

    mock_save_config.assert_called_once()
    saved_config, saved_path = mock_save_config.call_args[0]
    assert saved_path == mocked_session_path
    assert saved_config == expected_config
    assert result.exit_code == 0
    assert 'Configuration saved.' in result.output



# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from enum import Enum

class SpackViewEnum(Enum):
    VIEW = ''
    WITHOUT_VIEW = '--without-view'


# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.
import errno
#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import os
import shutil
import subprocess
import tempfile
from pathlib import Path

from oras.logger import Logger

from dedal.error_handling.exceptions import BashCommandException
import re


def clean_up(dirs: list[str], logger: logging = logging.getLogger(__name__), ignore_errors=True):
    """
        All the folders from the list dirs are removed with all the content in them
    """
    for cleanup_dir in dirs:
        cleanup_dir = Path(cleanup_dir).resolve()
        if cleanup_dir.exists():
            logger.info(f"Removing {cleanup_dir}")
            try:
                shutil.rmtree(Path(cleanup_dir))
            except OSError as e:
                logger.error(f"Failed to remove {cleanup_dir}: {e}")
                if not ignore_errors:
                    raise e
        else:
            logger.info(f"{cleanup_dir} does not exist")


def run_command(*args, logger=logging.getLogger(__name__), info_msg: str = '', exception_msg: str = None,
                exception=None, **kwargs):
    try:
        logger.info(f'{info_msg}: args: {args}')
        return subprocess.run(args, **kwargs)
    except subprocess.CalledProcessError as e:
        if exception_msg is not None:
            logger.error(f"{exception_msg}: {e}")
        if exception is not None:
            raise exception(f'{exception_msg} : {e}')
        else:
            return None
    except FileNotFoundError:
        logger.error(f"Command not found. Please check the command syntax.")
    except PermissionError:
        logger.error(f"Permission denied. Try running with appropriate permissions.")
    except subprocess.TimeoutExpired:
        logger.error(f"Command timed out. Try increasing the timeout duration.")
    except ValueError:
        logger.error(f"Invalid argument passed to subprocess. Check function parameters.")
    except OSError as e:
        logger.error(f"OS error occurred: {e}")
    except Exception as e:
        logger.error(f"An unexpected error occurred: {e}")
    return None


def git_clone_repo(repo_name: str, dir: Path, git_path: str, git_branch: str = '',
                   logger: Logger = logging.getLogger(__name__)):
    if not dir.exists():
        run_command(
            'git', f'clone', '--depth', '1',
            '-c', 'advice.detachedHead=false',
            '-c', 'feature.manyFiles=true',
            *(['-b', git_branch] if git_branch else []),
            git_path, dir
            , check=True, logger=logger,
            exception_msg=f'Failed to clone repository: {repo_name}',
            exception=BashCommandException)
    else:
        logger.info(f'Repository {repo_name} already cloned.')


def file_exists_and_not_empty(file: Path) -> bool:
    return file.is_file() and file.stat().st_size > 0


def log_command(results, log_file: str):
    try:
        with open(log_file, "w") as log_file:
            log_file.write(results.stdout)
            log_file.write("\n--- STDERR ---\n")
            log_file.write(results.stderr)
    except OSError as e:
        if e.errno == errno.ENOSPC:
            print(f"ERROR: No space left on device when writing")
        else:
            raise


def copy_to_tmp(file_path: Path) -> Path:
    """
    Creates a temporary directory and copies the given file into it.

    :param file_path: Path to the file that needs to be copied.
    :return: Path to the copied file inside the temporary directory.
    """
    if not file_path.is_file():
        raise FileNotFoundError(f"File not found: {file_path}")
    tmp_dir = Path(tempfile.mkdtemp())
    tmp_file_path = tmp_dir / file_path.name
    shutil.copy(file_path, tmp_file_path)
    return tmp_file_path


def set_bashrc_variable(var_name: str, value: str, bashrc_path: str = os.path.expanduser("~/.bashrc"),
                        logger: logging = logging.getLogger(__name__), update_variable=False):
    """Update or add an environment variable in ~/.bashrc."""
    if bashrc_path is None or var_name is None or value is None:
        return
    value = value.replace("$", r"\$")
    with open(bashrc_path, "r") as file:
        lines = file.readlines()
    pattern = re.compile(rf'^\s*export\s+{var_name}=.*$')
    found_variable = False
    # Modify the existing variable if found
    for i, line in enumerate(lines):
        if pattern.match(line):
            if update_variable:
                lines[i] = f'export {var_name}={value}\n'
            found_variable = True
            break
    if not found_variable:
        lines.append(f'\nexport {var_name}={value}\n')
        logger.info(f"Added in {bashrc_path} with: export {var_name}={value}")
    else:
        logger.info(f"Updated {bashrc_path} with: export {var_name}={value}")
    with open(bashrc_path, "w") as file:
        file.writelines(lines)


def copy_file(src: Path, dst: Path, logger: logging = logging.getLogger(__name__)) -> None:
    """
    Copy a file from src to dest.
    """
    if not os.path.exists(src):
        raise FileNotFoundError(f"Source file '{src}' does not exist.")
    src.resolve().as_posix()
    dst.resolve().as_posix()
    os.makedirs(os.path.dirname(dst), exist_ok=True)
    shutil.copy2(src, dst)
    logger.debug(f"File copied from '{src}' to '{dst}'")


def delete_file(file_path: str, logger: logging = logging.getLogger(__name__)) -> bool:
    """
    Deletes a file at the given path. Returns True if successful, False if the file doesn't exist.
    """
    try:
        os.remove(file_path)
        logger.debug(f"File '{file_path}' deleted.")
        return True
    except FileNotFoundError:
        logger.error(f"File not found: {file_path}")
        return False
    except PermissionError:
        logger.error(f"Permission denied: {file_path}")
        return False
    except Exception as e:
        logger.error(f"Error deleting file {file_path}: {e}")
        return False


def resolve_path(path: str):
    if path is None:
        path = Path(os.getcwd()).resolve()
    else:
        path = Path(path).resolve()
    return path


def count_files_in_folder(folder_path: Path) -> int:
    if not folder_path.is_dir():
        return 0
    return sum(1 for sub_path in folder_path.rglob("*") if sub_path.is_file())


def get_first_word(s: str) -> str:
    return s.split()[0] if s.strip() else ''



# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from collections.abc import Iterable
import pathlib
import ruamel.yaml as yaml
import spack

# migration of /cli/update_cached_buildresults.py
class SpecfileDagHash:
    def __init__(self, path_specfile, include_installed=False):
        self.path_specfile = pathlib.Path(path_specfile)
        self.include_installed = include_installed
        self.to_be_fetched = set()

    def run(self):
        with open(self.path_specfile, "r") as fd:
            file_content = fd.read()
            data = list(yaml.safe_load_all(file_content))

        for rspec in data:
            s = spack.spec.Spec.from_dict(rspec)
            if not isinstance(s, Iterable):
                s = [s]

            maybe_to_be_fetched = spack.traverse.traverse_nodes(s, key=spack.traverse.by_dag_hash)

            for spec in maybe_to_be_fetched:
                if not self.include_installed and spec.installed:
                    continue
                self.to_be_fetched.add(spec.dag_hash())
        to_be_fetched_dag_hash = []
        for dag_hash in self.to_be_fetched:
            to_be_fetched_dag_hash.append(dag_hash)
        return to_be_fetched_dag_hash


# spack-python script that merges two environment configuration files (spack.yaml) into one
# Usage: spack-python /path/to/first/spack.yaml /path/to/second/spack.yaml
# (note: if the second file does not exist, the output is the first file

import sys, os
from spack.config import merge_yaml, read_config_file, syaml

if not os.path.exists(sys.argv[2]):
    merged = syaml.dump(read_config_file(sys.argv[1]))
else:
    merged = syaml.dump(merge_yaml(read_config_file(sys.argv[1]), read_config_file(sys.argv[2])))

print(merged)




from abc import ABC, abstractmethod
from typing import Any

class SerializerInterface(ABC):
    @abstractmethod
    def serialize(self, obj: Any) -> None:
        """
        Serialize `obj` to the given file path.
        """
        pass

    @abstractmethod
    def deserialize(self) -> Any:
        """
        Deserialize and return the object from the given file path.
        """
        pass



# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from dedal.utils.utils import set_bashrc_variable


def test_add_new_variable(tmp_path):
    var_name = 'TEST_VAR'
    value = 'test_value'
    bashrc_path = tmp_path / ".bashrc"
    bashrc_path.touch()
    set_bashrc_variable(var_name, value, bashrc_path=str(bashrc_path))
    content = bashrc_path.read_text()
    assert f'export {var_name}={value}' in content


def test_update_existing_variable(tmp_path):
    var_name = 'TEST_VAR'
    value = 'test_value'
    updated_value = 'new_value'
    bashrc_path = tmp_path / ".bashrc"
    bashrc_path.write_text(f'export {var_name}={value}\n')
    set_bashrc_variable(var_name, updated_value, bashrc_path=str(bashrc_path), update_variable=True)
    content = bashrc_path.read_text()
    assert f'export {var_name}={updated_value}' in content
    assert f'export {var_name}={value}' not in content


def test_do_not_update_existing_variable(tmp_path):
    var_name = 'TEST_VAR'
    value = 'test_value'
    new_value = 'new_value'
    bashrc_path = tmp_path / ".bashrc"
    bashrc_path.write_text(f'export {var_name}={value}\n')

    set_bashrc_variable(var_name, new_value, bashrc_path=str(bashrc_path), update_variable=False)

    content = bashrc_path.read_text()
    assert f'export {var_name}={value}' in content
    assert f'export {var_name}={new_value}' not in content


def test_add_variable_with_special_characters(tmp_path):
    var_name = 'TEST_VAR'
    value = 'value_with_$pecial_chars'
    escaped_value = 'value_with_\\$pecial_chars'
    bashrc_path = tmp_path / ".bashrc"
    bashrc_path.touch()

    set_bashrc_variable(var_name, value, bashrc_path=str(bashrc_path))

    content = bashrc_path.read_text()
    assert f'export {var_name}={escaped_value}' in content



# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import argparse
import os
import pathlib
import subprocess

parser = argparse.ArgumentParser(
        prog='fetch_cached_buildresults.py',
        description='Downloading missing source files to a spack cache.',
        epilog='...')

parser.add_argument(
    "path_missing", type=pathlib.Path,
    help="Location of the output file that will list the packages not yet in the build cache.")

parser.add_argument(
    "specfiles", nargs="+",
    help="Location of the file containing the specs to be available.")

parser.add_argument(
    "--remote-cache", type=str, required=(not (("HARBOR_HOST" in os.environ) and ("HARBOR_PROJECT" in os.environ))),
    default=(("HARBOR_HOST" in os.environ) and ("HARBOR_PROJECT" in os.environ)) and
            ("{}/{}/build_cache".format(os.environ["HARBOR_HOST"], os.environ["HARBOR_PROJECT"])) or "",
    help="Path or URL to remote cache (target).")

parser.add_argument(
    "--remote-cache-type", type=str, choices=["oci"],
    default="oci",
    help="Type of the remote cache.")

parser.add_argument(
    "--remote-cache-username", type=str,
    default="HARBOR_USERNAME" in os.environ and
        pathlib.Path(os.environ["HARBOR_USERNAME"]) or "",
    help="Username for remote cache (if applicable)")

parser.add_argument(
    "--remote-cache-password", type=str,
    default="HARBOR_PASSWORD" in os.environ and
        pathlib.Path(os.environ["HARBOR_PASSWORD"]) or None,
    help="Password for remote cache (if applicable)")

parser.add_argument(
    "--local-cache", type=str,
    default="YASHCHIKI_CACHE_BUILD" in os.environ and
        pathlib.Path(os.environ["YASHCHIKI_CACHE_BUILD"]) or
        os.path.expanduser("~/.yashchiki/cache/"),
    help="Path to local spack cache folder (build results).")

parser.add_argument(
    "--yashchiki-home", type=str, required=True,
    help="Path to yashchiki home for calling helper tools.")

parser.add_argument(
    "--include-installed",
    action='store_true', default=False,
    help="Include already installed specs.")

args = parser.parse_args()

local_cache = pathlib.Path(args.local_cache)
if not os.path.exists(args.local_cache):
    print("Creating local build cache directory")
    local_cache.mkdir(parents=True, exist_ok=False)

missing_packages = []
available_packages = []
cached_paths = []
cmd = ["oras", "repo", "tags"]
if args.remote_cache_username and args.remote_cache_password:
    cmd.extend(["--username", args.remote_cache_username])
    cmd.extend(["--password", args.remote_cache_password])
cmd.append(args.remote_cache)
try:
    tags = subprocess.check_output(cmd)
    tags = tags.decode("utf-8")
    cached_paths = tags.split()
except subprocess.CalledProcessError as e:
    print(f"Listing repo tags of \"{args.remote_cache}\" failed.")

for specfile in args.specfiles:
    with open(specfile, "r") as fd:
        fetch_paths = []
        packages = {}
        try:
            include_installed = " --include-installed" if args.include_installed else ""
            # FIXME: import and call function, but this would need *this to be run in spack-python already
            lines = subprocess.check_output(f"spack-python {args.yashchiki_home}/specfile_storage_path_build.py {specfile}{include_installed}", shell=True)
            lines = lines.decode("utf-8")
            lines = lines.split("\n")
            for line in lines:
                if not line:
                    continue
                elems = line.split()
                packages[elems[0]] = elems[1:]
        except subprocess.CalledProcessError as e:
            print(f"Computing fetch buildresult paths failed:", str(e), e.output)
        for package_dag_hash, fetch_paths in packages.items():
            missing_paths = []
            for fetch_path in fetch_paths:
                basename = os.path.basename(fetch_path)
                if basename in cached_paths:
                    cmd = ["oras", "pull"]
                    if args.remote_cache_username and args.remote_cache_password:
                        cmd.extend(["--username", args.remote_cache_username])
                        cmd.extend(["--password", args.remote_cache_password])
                    cmd.append(args.remote_cache + f":{basename}")
                    try:
                        subprocess.check_output(cmd, stderr=subprocess.STDOUT, cwd=local_cache)
                    except subprocess.CalledProcessError as e:
                        print(f"Pulling of \"{basename}\" from \"{args.remote_cache}\" failed.")
                        missing_paths.append(fetch_path)
                else:
                    missing_paths.append(fetch_path)
            package_missing = False
            for missing_path in missing_paths:
                if missing_path.endswith(".spack") or missing_path.endswith(".spec.json"):
                    package_missing = True
            if package_missing:
                missing_packages.append(f"{package_dag_hash} " + " ".join(missing_paths))
            else:
                available_packages.append(f"{package_dag_hash} " + " ".join(missing_paths))

print(len(missing_packages), "missing packages in remote buildresults cache.")
print(len(available_packages), "available packages in remote buildresults cache.")

if missing_packages:
    with open(args.path_missing, "w") as fd:
        fd.write("\n".join(missing_packages))



# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import subprocess
from pathlib import Path

from dedal.build_cache.BuildCacheManagerCreator import BuildCacheManagerCreator
from dedal.configuration.SpackConfig import SpackConfig
from dedal.error_handling.exceptions import SpackInstallPackagesException
from dedal.logger.logger_builder import get_logger
from dedal.spack_factory.SpackOperation import SpackOperation
from dedal.utils.utils import file_exists_and_not_empty, run_command, log_command, copy_file
from dedal.wrapper.spack_wrapper import check_spack_env


class SpackOperationUseCache(SpackOperation):
    """
    This class uses caching for the concretization step and for the installation step.
    """

    def __init__(self, spack_config: SpackConfig = SpackConfig()):
        super().__init__(spack_config, logger=get_logger(__name__))
        self.signed = False
        self.cache_dependency, self.build_cache = BuildCacheManagerCreator.get_build_cache_manager_operator(
            spack_config)

    def setup_spack_env(self) -> None:
        """Set up the spack environment for using the cache.
        Downloads the build cache, adds the public key to trusted keys,
        and adds the build cache mirror.
        Raises:
            NoSpackEnvironmentException: If the spack environment is not set up.
        """
        super().setup_spack_env()
        # Download concretization cache from OCI Registry
        if self.cache_dependency:
            self.cache_dependency.download(self.spack_config.concretization_dir)
        # Download build cache from OCI Registry and add public key to trusted keys
        if self.build_cache:
            self.build_cache.download(self.spack_config.buildcache_dir)
            cached_public_key = self.build_cache.get_public_key_from_cache(str(self.spack_config.buildcache_dir))
            self.signed = cached_public_key is not None
            if self.signed:
                self.trust_gpg_key(cached_public_key)
            # Add build cache mirror
            self.add_mirror('local_cache',
                            str(self.spack_config.buildcache_dir),
                            signed=self.signed,
                            autopush=False,
                            global_mirror=False)
            self.update_buildcache_index(self.spack_config.buildcache_dir)
            self.install_gpg_keys()

    @check_spack_env
    def concretize_spack_env(self, test=None):
        """Concretization step for spack environment for using the concretization cache (spack.lock file).
        Downloads the concretization cache and moves it to the spack environment's folder
        Raises:
            NoSpackEnvironmentException: If the spack environment is not set up.
        """
        concretization_redo = False
        if self.cache_dependency and file_exists_and_not_empty(self.spack_config.concretization_dir / 'spack.lock'):
            concretization_file_path = self.env_path / 'spack.lock'
            copy_file(self.spack_config.concretization_dir / 'spack.lock', self.env_path)
            # redo the concretization step if spack.lock file was not downloaded from the cache
            if not file_exists_and_not_empty(concretization_file_path):
                super().concretize_spack_env(force=True, test=test)
                concretization_redo = True
        else:
            # redo the concretization step if spack.lock file was not downloaded from the cache
            super().concretize_spack_env(force=True, test=test)
            concretization_redo = True
        if concretization_redo is True:
            self.logger.info(f'Redo of concretization step.')
        else:
            self.logger.info(f'Used concretization from cache.')
        return concretization_redo

    @check_spack_env
    def install_packages(self, jobs: int, debug=False, test=None):
        """Installation step for spack environment for using the binary caches.

        Raises:
            NoSpackEnvironmentException: If the spack environment is not set up.
        """
        signed = '' if self.signed else '--no-check-signature'
        debug = '--debug' if debug else ''
        test = f'--test {test}' if test else ''
        if self.build_cache:
            install_result = run_command("bash", "-c",
                                         f'{self.spack_command_on_env} && spack {debug} install {signed} -j {jobs} {test}',
                                         stdout=subprocess.PIPE,
                                         stderr=subprocess.PIPE,
                                         text=True,
                                         logger=self.logger,
                                         info_msg=f"Installing spack packages for {self.spack_config.env.name}",
                                         exception_msg=f"Error installing spack packages for {self.spack_config.env.name}",
                                         exception=SpackInstallPackagesException)
            log_command(install_result, str(Path(os.getcwd()).resolve() / ".generate_cache.log"))
            if install_result.returncode == 0:
                self.logger.info(f'Finished installation of spack packages from cache.')
            else:
                self.logger.error(f'Something went wrong during installation from cache. Please check the logs.')
        else:
            install_result = super().install_packages(jobs=jobs, signed=signed, debug=debug, test=test)
        return install_result



# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import pathlib
import subprocess
from typing import List

from dedal.build_cache.BuildCacheManagerOciDefault import BuildCacheManagerOciDefault
from dedal.build_cache.SpecfileBuildStoragePathExtractor import SpecfileBuildStoragePathExtractor
from dedal.logger.logger_builder import get_logger

default_local_cache = "YASHCHIKI_CACHE_BUILD" in os.environ and pathlib.Path(
    os.environ["YASHCHIKI_CACHE_BUILD"]) or os.path.expanduser("~/.dedal/cache/")
default_remote_cache = ("HARBOR_HOST" in os.environ) and ("HARBOR_PROJECT" in os.environ) and (
    "{}/{}/build_cache".format(os.environ["HARBOR_HOST"], os.environ["HARBOR_PROJECT"])) or ""
default_remote_cache_username = "HARBOR_USERNAME" in os.environ and pathlib.Path(os.environ["HARBOR_USERNAME"]) or ""
default_remote_cache_password = "HARBOR_PASSWORD" in os.environ and pathlib.Path(os.environ["HARBOR_PASSWORD"]) or None


# migration of /cli/fetch_cached_buildresults.py
class CachedBuildResultsFetcher:
    def __init__(self, path_missing: pathlib.Path, specfiles: List[str], dedal_home: str,
                 remote_cache: str = default_remote_cache,
                 remote_cache_type: str = "oci", remote_cache_username: str = default_remote_cache_username,
                 remote_cache_password: str = default_remote_cache_password,
                 local_cache: str = default_local_cache, include_installed: bool = False):
        self.path_missing = pathlib.Path(path_missing)
        self.specfiles = specfiles
        self.dedal_home = dedal_home
        self.remote_cache = remote_cache
        self.remote_cache_type = remote_cache_type
        self.remote_cache_username = remote_cache_username
        self.remote_cache_password = remote_cache_password
        self.local_cache = pathlib.Path(local_cache)
        self.include_installed = include_installed
        self.logger = get_logger(__name__)

        if self.local_cache.exists():
            self.logger.info("Creating local build cache directory")
            self.local_cache.mkdir(parents=True, exist_ok=False)

        host, project, _ = self.remote_cache.split("/", 2)
        self.oci = BuildCacheManagerOciDefault(
            registry_host=host,
            registry_project=project,
            registry_username=self.remote_cache_username,
            registry_password=self.remote_cache_password,
        )

    def fetch(self):
        cached_paths = self.oci.list_tags()
        missing_packages = []
        available_packages = []

        for specfile in self.specfiles:
            with open(specfile, "r") as fd:
                try:
                    extractor = SpecfileBuildStoragePathExtractor(include_installed=self.include_installed)
                    lines = extractor.extract_paths(pathlib.Path(specfile))
                    packages = {line.split()[0]: line.split()[1:] for line in lines}
                except subprocess.CalledProcessError as e:
                    self.logger.info(f"Computing fetch buildresult paths failed: {str(e)}")
                    continue

                for package_dag_hash, fetch_paths in packages.items():
                    missing_paths = []
                    for fetch_path in fetch_paths:
                        basename = os.path.basename(fetch_path)
                        if basename in cached_paths:
                            try:
                                self.oci.client.pull(
                                    ref=f"{self.remote_cache}:{basename}",
                                    outdir=str(self.local_cache),
                                    overwrite=True
                                )
                            except Exception as e:
                                self.logger.info(f"Pulling of \"{basename}\" from \"{self.remote_cache}\" failed.")
                                missing_paths.append(fetch_path)
                        else:
                            missing_paths.append(fetch_path)

                    if any(p.endswith((".spack", ".spec.json")) for p in missing_paths):
                        missing_packages.append(f"{package_dag_hash} " + " ".join(missing_paths))
                    else:
                        available_packages.append(f"{package_dag_hash} " + " ".join(missing_paths))

        self.logger.info(len(missing_packages), "missing packages in remote buildresults cache.")
        self.logger.info(len(available_packages), "available packages in remote buildresults cache.")

        if missing_packages:
            with open(self.path_missing, "w") as fd:
                fd.write("\n".join(missing_packages))



# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
from pathlib import Path
import click
import jsonpickle

from dedal.bll.SpackManager import SpackManager
from dedal.bll.cli_utils import save_config, load_config
from dedal.configuration.GpgConfig import GpgConfig
from dedal.configuration.SpackConfig import SpackConfig
from dedal.enum.SpackViewEnum import SpackViewEnum
from dedal.model.SpackDescriptor import SpackDescriptor
from dedal.utils.utils import resolve_path

SESSION_CONFIG_PATH = os.path.expanduser('/tmp/dedal/dedal_session.json')
os.makedirs(os.path.dirname(SESSION_CONFIG_PATH), exist_ok=True)


@click.group()
@click.pass_context
def cli(ctx: click.Context):
    config = load_config(SESSION_CONFIG_PATH)
    if ctx.invoked_subcommand not in ['set-config', 'install-spack'] and not config:
        click.echo('No configuration set. Use `set-config` first.')
        ctx.exit(1)
    if config:
        config['env_path'] = resolve_path(config['env_path'])
        env = SpackDescriptor(config['env_name'], config['env_path'], config['env_git_path'])
        gpg = GpgConfig(config['gpg_name'], config['gpg_mail']) if config['gpg_name'] and config['gpg_mail'] else None
        spack_config = SpackConfig(env=env, repos=None, install_dir=config['install_dir'],
                                   upstream_instance=config['upstream_instance'],
                                   concretization_dir=config['concretization_dir'],
                                   buildcache_dir=config['buildcache_dir'],
                                   system_name=config['system_name'], gpg=gpg,
                                   use_spack_global=config['use_spack_global'],
                                   view=config['view'],
                                   override_cache=config['override_cache'])
        ctx.obj = SpackManager(spack_config, use_cache=config['use_cache'])


@cli.command()
@click.option('--use_cache', is_flag=True, default=False, help='Enables cashing')
@click.option('--use_spack_global', is_flag=True, default=False, help='Uses spack installed globally on the os')
@click.option('--env_name', type=str, default=None, help='Environment name')
@click.option('--env_path', type=str, default=None, help='Environment path to download locally')
@click.option('--env_git_path', type=str, default=None, help='Git path to download the environment')
@click.option('--install_dir', type=str,
              help='Install directory for installing spack; spack environments and repositories are stored here')
@click.option('--upstream_instance', type=str, default=None, help='Upstream instance for spack environment')
@click.option('--system_name', type=str, default=None, help='System name; it is used inside the spack environment')
@click.option('--concretization_dir', type=str, default=None,
              help='Directory where the concretization caching (spack.lock) will be downloaded')
@click.option('--buildcache_dir', type=str, default=None,
              help='Directory where the binary caching is downloaded for the spack packages')
@click.option('--gpg_name', type=str, default=None, help='Gpg name')
@click.option('--gpg_mail', type=str, default=None, help='Gpg mail contact address')
@click.option('--cache_version_concretize', type=str, default='v1', help='Cache version for concretizaion data')
@click.option('--cache_version_build', type=str, default='v1', help='Cache version for binary caches data')
@click.option('--view', type=SpackViewEnum, default=SpackViewEnum.VIEW, help='Spack environment view')
@click.option('--override_cache', is_flag=True, default=True, help='Flag for overriding existing cache')
def set_config(use_cache, env_name, env_path, env_git_path, install_dir, upstream_instance, system_name,
               concretization_dir,
               buildcache_dir, gpg_name, gpg_mail, use_spack_global, cache_version_concretize, cache_version_build,
               view, override_cache):
    """Sets configuration parameters for the session."""
    spack_config_data = {
        'use_cache': use_cache,
        'env_name': env_name,
        'env_path': env_path,
        'env_git_path': env_git_path,
        'install_dir': install_dir,
        'upstream_instance': upstream_instance,
        'system_name': system_name,
        'concretization_dir': Path(concretization_dir) if concretization_dir else None,
        'buildcache_dir': Path(buildcache_dir) if buildcache_dir else None,
        'gpg_name': gpg_name,
        'gpg_mail': gpg_mail,
        'use_spack_global': use_spack_global,
        'repos': [],
        'cache_version_concretize': cache_version_concretize,
        'cache_version_build': cache_version_build,
        'view': view,
        'override_cache': override_cache,
    }
    save_config(spack_config_data, SESSION_CONFIG_PATH)
    click.echo('Configuration saved.')


@click.command()
def show_config():
    """Show the current configuration."""
    config = load_config(SESSION_CONFIG_PATH)
    if config:
        click.echo(jsonpickle.encode(config, indent=2))
    else:
        click.echo('No configuration set. Use `set-config` first.')


@cli.command()
@click.option('--spack_version', type=str, default='0.23.0',
              help='Specifies the Spack version to be installed (default: v0.23.0).')
@click.option('--bashrc_path', type=str, default="~/.bashrc", help='Defines the path to .bashrc.')
@click.pass_context
def install_spack(ctx: click.Context, spack_version: str, bashrc_path: str):
    """Install spack in the install_dir folder"""
    bashrc_path = os.path.expanduser(bashrc_path)
    if ctx.obj is None:
        SpackManager().install_spack(spack_version, bashrc_path)
    else:
        ctx.obj.install_spack(spack_version, bashrc_path)


@cli.command()
@click.option('--repo_name', type=str, required=True, default=None, help='Repository name')
@click.option('--path', type=str, required=True, default=None, help='Repository path to download locally')
@click.option('--git_path', type=str, required=True, default=None, help='Git path to download the repository')
def add_spack_repo(repo_name: str, path: str, git_path: str = None):
    """Adds a spack repository to the spack environments. The setup command must be rerun."""
    path = resolve_path(path)
    repo = SpackDescriptor(repo_name, path, git_path)
    config = load_config(SESSION_CONFIG_PATH)
    config['repos'].append(repo)
    save_config(config, SESSION_CONFIG_PATH)
    click.echo('dedal setup_spack_env must be reran after each repo is added for the environment.')


@cli.command()
@click.pass_context
def setup_spack_env(ctx: click.Context):
    """Setups a spack environment according to the given configuration."""
    ctx.obj.setup_spack_env()


@cli.command()
@click.pass_context
def concretize(ctx: click.Context):
    """Spack concretization step."""
    ctx.obj.concretize_spack_env()


@cli.command()
@click.option('--jobs', type=int, default=2, help='Number of parallel jobs for spack installation')
@click.pass_context
def install_packages(ctx: click.Context, jobs):
    """Installs spack packages present in the spack environment defined in configuration."""
    ctx.obj.install_packages(jobs=jobs)


@click.command()
def clear_config():
    """Clears stored configuration."""
    if os.path.exists(SESSION_CONFIG_PATH):
        os.remove(SESSION_CONFIG_PATH)
        click.echo('Configuration cleared!')
    else:
        click.echo('No configuration to clear.')


cli.add_command(show_config)
cli.add_command(clear_config)

if __name__ == '__main__':
    cli()



# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import builtins
import oras.client
from dedal.logger.logger_builder import get_logger


class BuildCacheManagerOciDefault:
    """
        This class aims to manage the push/pull/delete of build cache files
    """

    def __new__(cls, registry_host, registry_project, registry_username, registry_password, cache_version='cache',
                auth_backend='basic', insecure=False, tls_verify=True):
        instance = super().__new__(cls)
        instance._logger = get_logger(__name__, BuildCacheManagerOciDefault.__name__)
        instance._registry_project = registry_project

        instance._registry_username = registry_username
        instance._registry_password = registry_password

        instance._registry_host = registry_host

        # Override input to disable prompts during login.
        # Define a function that raises an exception when input is called.
        def disabled_input():
            raise Exception("Interactive login disabled: credentials are provided via attributes.")

        # Save the original input function.
        original_input = builtins.input
        # Override input so that any call to input() during login will raise our exception.
        builtins.input = disabled_input
        # Initialize an OrasClient instance.
        # This method utilizes the OCI Registry for container image and artifact management.
        # Refer to the official OCI Registry documentation for detailed information on the available authentication methods.
        # Supported authentication types may include basic authentication (username/password), token-based authentication,
        instance.client = oras.client.OrasClient(hostname=instance._registry_host, auth_backend=auth_backend,
                                                 insecure=insecure, tls_verify=tls_verify)

        try:
            instance.client.login(username=instance._registry_username, password=instance._registry_password)
        except Exception:
            instance._logger.error('Login failed!')
            return None
        finally:
            builtins.input = original_input

        instance.cache_version = cache_version
        instance._oci_registry_path = f'{instance._registry_host}/{instance._registry_project}/{instance.cache_version}'
        return instance

    def list_tags(self):
        """
            This method retrieves all tags from an OCI Registry
        """
        try:
            return self.client.get_tags(self._oci_registry_path)
        except Exception as e:
            self._logger.error(f"Failed to list tags: {e}")
            return []


# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

from dedal.configuration.SpackConfig import SpackConfig
from dedal.spack_factory.SpackOperation import SpackOperation
from dedal.spack_factory.SpackOperationCreateCache import SpackOperationCreateCache
from dedal.spack_factory.SpackOperationUseCache import SpackOperationUseCache


class SpackOperationCreator:
    @staticmethod
    def get_spack_operator(spack_config: SpackConfig = None, use_cache: bool = False) -> SpackOperation:
        env_vars_concretization_cache = [
            os.environ.get('CONCRETIZE_OCI_HOST'),
            os.environ.get('CONCRETIZE_OCI_PROJECT'),
            os.environ.get('CONCRETIZE_OCI_USERNAME'),
            os.environ.get('CONCRETIZE_OCI_PASSWORD'),
        ]
        env_vars_build_cache = [
            os.environ.get('BUILDCACHE_OCI_HOST'),
            os.environ.get('BUILDCACHE_OCI_PROJECT'),
            os.environ.get('BUILDCACHE_OCI_USERNAME'),
            os.environ.get('BUILDCACHE_OCI_PASSWORD')
        ]
        if spack_config is None:
            return SpackOperation()
        elif None in env_vars_concretization_cache and None in env_vars_build_cache:
            return SpackOperation(spack_config)
        elif spack_config.concretization_dir is None and spack_config.buildcache_dir is None:
            return SpackOperation(spack_config)
        elif (spack_config.concretization_dir or spack_config.buildcache_dir) and not use_cache:
            return SpackOperationCreateCache(spack_config)
        elif (spack_config.concretization_dir or spack_config.buildcache_dir) and use_cache:
            return SpackOperationUseCache(spack_config)
        else:
            return SpackOperation(SpackConfig())



# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import glob
import os
from os.path import join
from pathlib import Path
from dedal.build_cache.BuildCacheManagerInterface import BuildCacheManagerInterface
from dedal.build_cache.BuildCacheManagerOciDefault import BuildCacheManagerOciDefault


class BuildCacheManagerOci(BuildCacheManagerOciDefault, BuildCacheManagerInterface):
    """
        This class aims to manage the push/pull/delete of build cache files
    """

    def upload(self, upload_dir: Path, override_cache=True):
        """
            This method pushed all the files from the build cache folder into the OCI Registry
            Args:
                upload_dir (Path): directory with the local binary caches
                override_cache (bool): Updates the cache from the OCI Registry with the same tag
        """
        build_cache_path = upload_dir.resolve()
        # build cache folder must exist before pushing all the artifacts
        if not build_cache_path.exists():
            self._logger.error(f"Path {build_cache_path} not found.")

        tags = self.list_tags()

        for sub_path in build_cache_path.rglob("*"):
            if sub_path.is_file():
                tag = str(sub_path.name)
                rel_path = str(sub_path.relative_to(build_cache_path)).replace(tag, "")
                target = f"{self._registry_host}/{self._registry_project}/{self.cache_version}:{tag}"
                upload_file = True
                if override_cache is False and tag in tags:
                    upload_file = False
                if upload_file:
                    try:
                        self._logger.info(f"Pushing file '{sub_path}' to ORAS target '{target}' ...")
                        self.client.push(
                            files=[str(sub_path)],
                            target=target,
                            # save in manifest the relative path for reconstruction
                            manifest_annotations={"path": rel_path},
                            disable_path_validation=True,
                        )
                        self._logger.info(f"Successfully pushed {tag}")
                    except Exception as e:
                        self._logger.error(
                            f"An error occurred while pushing: {e}")
                else:
                    self._logger.info(f"File '{sub_path}' already uploaded ...")

    def download(self, download_dir: Path):
        """
            This method pulls all the files from the OCI Registry into the build cache folder
        """
        build_cache_path = download_dir.resolve()
        # create the buildcache dir if it does not exist
        os.makedirs(build_cache_path, exist_ok=True)
        tags = self.list_tags()
        if tags is not None:
            for tag in tags:
                ref = f"{self._registry_host}/{self._registry_project}/{self.cache_version}:{tag}"
                # reconstruct the relative path of each artifact by getting it from the manifest
                cache_path = \
                    self.client.get_manifest(
                        f'{self._registry_host}/{self._registry_project}/{self.cache_version}:{tag}')[
                        'annotations'][
                        'path']
                try:
                    self.client.pull(
                        ref,
                        outdir=str(build_cache_path / cache_path),
                        overwrite=True
                    )
                    self._logger.info(f"Successfully pulled artifact {tag}.")
                except Exception as e:
                    self._logger.error(
                        f"Failed to pull artifact {tag} : {e}")

    def delete(self):
        """
            Deletes all artifacts from an OCI Registry based on their tags.
            This method removes artifacts identified by their tags in the specified OCI Registry.
            It requires appropriate permissions to delete artifacts from the registry.
            If the registry or user does not have the necessary delete permissions, the operation might fail.
        """
        tags = self.list_tags()
        if tags is not None:
            try:
                self.client.delete_tags(self._oci_registry_path, tags)
                self._logger.info("Successfully deleted all artifacts form OCI registry.")
            except RuntimeError as e:
                self._logger.error(
                    f"Failed to delete artifacts: {e}")

    def __log_warning_if_needed(self, warn_message: str, items: list[str]) -> None:
        """Logs a warning message if the number of items is greater than 1. (Private function)
           This method logs a warning message using the provided message and items if the list of items has more than one element.

        Args:
            warn_message (str): The warning message to log.
            items (list[str]): The list of items to include in the log message.
        """
        if len(items) > 1:
            self._logger.warning(warn_message, items, items[0])

    def get_public_key_from_cache(self, build_cache_dir: str | None) -> str | None:
        """Retrieves the public key from the build cache.
            This method searches for the public key within the specified build cache directory.
        Args:
            build_cache_dir (str | None): The path to the build cache directory.
        Returns:
            str | None: The path to the public key file if found, otherwise None.
        """

        if not build_cache_dir or not os.path.exists(build_cache_dir):
            self._logger.warning("Build cache directory does not exist!")
            return None
        pgp_folders = glob.glob(f"{build_cache_dir}/**/_pgp", recursive=True)
        if not pgp_folders:
            self._logger.warning("No _pgp folder found in the build cache!")
            return None
        self.__log_warning_if_needed(
            "More than one PGP folders found in the build cache: %s, using the first one in the list: %s", pgp_folders)
        pgp_folder = pgp_folders[0]
        key_files = glob.glob(join(pgp_folder, "**"))
        if not key_files:
            self._logger.warning("No PGP key files found in the build cache!")
            return None
        self.__log_warning_if_needed(
            "More than one PGP key files found in the build cache: %s, using the first one in the list: %s", key_files)
        return key_files[0]



# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

from dedal.spack_factory.SpackOperationCreateCache import SpackOperationCreateCache

from dedal.configuration.SpackConfig import SpackConfig
from dedal.model.SpackDescriptor import SpackDescriptor
from dedal.spack_factory.SpackOperation import SpackOperation
from dedal.spack_factory.SpackOperationCreator import SpackOperationCreator
from dedal.spack_factory.SpackOperationUseCache import SpackOperationUseCache
from dedal.tests.testing_variables import ebrains_spack_builds_git, test_spack_env_git


def test_spack_creator_scratch_1(tmp_path):
    install_dir = tmp_path
    env = SpackDescriptor('test-spack-env', install_dir, test_spack_env_git)
    repo = SpackDescriptor('ebrains-spack-builds', install_dir, ebrains_spack_builds_git)
    spack_config = SpackConfig(env, install_dir=install_dir)
    spack_config.add_repo(repo)
    spack_operation = SpackOperationCreator.get_spack_operator(spack_config)
    assert isinstance(spack_operation, SpackOperation)


def test_spack_creator_scratch_2(tmp_path):
    spack_config = None
    spack_operation = SpackOperationCreator.get_spack_operator(spack_config)
    assert isinstance(spack_operation, SpackOperation)


def test_spack_creator_scratch_3():
    spack_config = SpackConfig()
    spack_operation = SpackOperationCreator.get_spack_operator(spack_config)
    assert isinstance(spack_operation, SpackOperation)

def test_spack_creator_scratch_4(tmp_path):
    concretize_oci_host = os.environ.get('CONCRETIZE_OCI_HOST')
    os.environ.pop('CONCRETIZE_OCI_HOST', None)
    install_dir = tmp_path
    env = SpackDescriptor('test-spack-env', install_dir, test_spack_env_git)
    repo = SpackDescriptor('ebrains-spack-builds', install_dir, ebrains_spack_builds_git)
    spack_config = SpackConfig(env, install_dir=install_dir, concretization_dir=install_dir, buildcache_dir=install_dir)
    spack_config.add_repo(repo)
    spack_operation = SpackOperationCreator.get_spack_operator(spack_config)
    os.environ['CONCRETIZE_OCI_HOST'] = concretize_oci_host
    assert isinstance(spack_operation, SpackOperation)


def test_spack_creator_create_cache(tmp_path):
    install_dir = tmp_path
    env = SpackDescriptor('test-spack-env', install_dir, test_spack_env_git)
    repo = SpackDescriptor('ebrains-spack-builds', install_dir, ebrains_spack_builds_git)
    spack_config = SpackConfig(env, install_dir=install_dir, concretization_dir=install_dir, buildcache_dir=install_dir)
    spack_config.add_repo(repo)
    spack_operation = SpackOperationCreator.get_spack_operator(spack_config)
    assert isinstance(spack_operation, SpackOperationCreateCache)


def test_spack_creator_use_cache(tmp_path):
    install_dir = tmp_path
    env = SpackDescriptor('test-spack-env', install_dir, test_spack_env_git)
    repo = SpackDescriptor('ebrains-spack-builds', install_dir, ebrains_spack_builds_git)
    spack_config = SpackConfig(env, install_dir=install_dir, concretization_dir=install_dir, buildcache_dir=install_dir)
    spack_config.add_repo(repo)
    spack_operation = SpackOperationCreator.get_spack_operator(spack_config, use_cache=True)
    assert isinstance(spack_operation, SpackOperationUseCache)



# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.
import os

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import yaml
import re


def extract_spack_packages(spack_yaml_path: str):
    """
    Extracts only the package names from a spack.yaml file, ignoring versions, variants, compilers, and dependencies.
    Args:
        spack_yaml_path (str): Path to the spack.yaml file.
    Returns:
        List[str]: Clean list of package names.
    """
    with open(spack_yaml_path, 'r') as f:
        data = yaml.safe_load(f)
    specs = data.get('spack', {}).get('specs', [])
    package_names = []
    for spec in specs:
        if isinstance(spec, str):
            # Match the first word-like token before any of: @ + % ^ whitespace
            match = re.match(r'^([\w-]+)', spec.strip())
            if match:
                package_names.append(match.group(1))

    return package_names


def find_first_upstream_prefix(upstream_instance):
    """
    Search for directories named '.spack-db' within the spack opt directory
    under upstream_instance, and return a list of their parent directories.
    """
    base_path = os.path.join(upstream_instance, "spack", "opt", "spack")
    for upstream_prefix, dirs, _ in os.walk(base_path):
        if ".spack-db" in dirs:
            return upstream_prefix
    return None



# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from collections.abc import Iterable
import pathlib
import ruamel.yaml as yaml
import spack
import spack.binary_distribution as bindist


# migration of /cli/specfile_storage_path_build.py
class SpecfileBuildStoragePathExtractor:
    def __init__(self, include_installed: bool = False):
        self.include_installed = include_installed

    def extract_paths(self, path_specfile: pathlib.Path):
        with open(path_specfile, "r") as fd:
            file_content = fd.read()
            data = list(yaml.safe_load_all(file_content))

        to_be_fetched = set()

        for rspec in data:
            specs = spack.spec.Spec.from_dict(rspec)
            if not isinstance(specs, Iterable):
                specs = [specs]

            for spec in spack.traverse.traverse_nodes(specs, key=spack.traverse.by_dag_hash):
                if not self.include_installed and spec.installed:
                    continue

                build_cache_paths = [
                    bindist.tarball_path_name(spec, ".spack"),
                    bindist.tarball_name(spec, ".spec.json.sig"),
                    bindist.tarball_name(spec, ".spec.json"),
                    bindist.tarball_name(spec, ".spec.yaml"),
                ]
                to_be_fetched.add(str(spec.dag_hash()) + " " + " ".join(build_cache_paths))

        return list(to_be_fetched)



# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
from pathlib import Path
from dedal.configuration.GpgConfig import GpgConfig
from dedal.enum.SpackCacheStorageEnum import SpackCacheStorageEnum
from dedal.enum.SpackViewEnum import SpackViewEnum
from dedal.model.SpackDescriptor import SpackDescriptor
from dedal.utils.utils import resolve_path


class SpackConfig:
    def __init__(self, env: SpackDescriptor = None, repos: list[SpackDescriptor] = None,
                 install_dir=Path(os.getcwd()).resolve(), upstream_instance=None, system_name=None,
                 concretization_dir: Path = None, buildcache_dir: Path = None, gpg: GpgConfig = None,
                 use_spack_global=False, cache_version_concretize='v1',
                 cache_version_build='v1', view=SpackViewEnum.VIEW, override_cache=True, bashrc_path=os.path.expanduser("~/.bashrc"),
                 spack_cache_storage_type: SpackCacheStorageEnum = SpackCacheStorageEnum.OCI,
                 spack_dir: Path = None, serialize_name: str = 'data.pkl'):
        self.env = env
        self.serialize_name = serialize_name
        if repos is None:
            self.repos = []
        else:
            self.repos = repos
        self.spack_dir = spack_dir
        self.upstream_instance = upstream_instance
        self.system_name = system_name
        self.concretization_dir = concretization_dir if concretization_dir is None else resolve_path(concretization_dir)
        self.buildcache_dir = buildcache_dir if buildcache_dir is None else resolve_path(buildcache_dir)
        self.install_dir = resolve_path(install_dir)
        self.gpg = gpg
        self.use_spack_global = use_spack_global
        self.cache_version_concretize = cache_version_concretize
        self.cache_version_build = cache_version_build
        self.view = view
        self.override_cache = override_cache
        self.bashrc_path = bashrc_path
        self.spack_cache_storage_type = spack_cache_storage_type

    def add_repo(self, repo: SpackDescriptor):
        if self.repos is None:
            self.repos = []
        else:
            self.repos.append(repo)



# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pathlib import Path

import pytest

from dedal.configuration.GpgConfig import GpgConfig
from dedal.configuration.SpackConfig import SpackConfig

from dedal.model.SpackDescriptor import SpackDescriptor
from dedal.spack_factory.SpackOperationCreateCache import SpackOperationCreateCache
from dedal.spack_factory.SpackOperationCreator import SpackOperationCreator
from dedal.tests.testing_variables import test_spack_env_git, ebrains_spack_builds_git

"""
Before running those tests, the repositories where the caching is stored must be cleared after each run. 
Ebrains Harbour does not support deletion via API, so the clean up must be done manually
"""


def setup(path):
    install_dir = path
    env_name = 'test-spack-env'
    env_path = install_dir / env_name
    concretization_dir = install_dir / 'concretization'
    buildcache_dir = install_dir / 'buildcache'
    env = SpackDescriptor(env_name, install_dir, test_spack_env_git)
    repo = SpackDescriptor('ebrains-spack-builds', install_dir, ebrains_spack_builds_git)
    config = SpackConfig(env=env, install_dir=install_dir, concretization_dir=concretization_dir,
                         buildcache_dir=buildcache_dir, gpg=None)
    config.add_repo(repo)
    spack_operation = SpackOperationCreator.get_spack_operator(config)
    return spack_operation, env_path


@pytest.mark.skip(
    reason="Skipping until an OCI registry which supports via API deletion; Clean up for OCI registry repo must be added before this test.")
def test_spack_create_cache_concretization(tmp_path):
    install_dir = tmp_path
    spack_operation, _ = setup(install_dir)
    assert isinstance(spack_operation, SpackOperationCreateCache)
    spack_operation.install_spack(bashrc_path=str(tmp_path / Path('.bashrc')))
    spack_operation.setup_spack_env()
    spack_operation.concretize_spack_env()
    assert len(spack_operation.cache_dependency.list_tags()) > 0


@pytest.mark.skip(
    reason="Skipping until an OCI registry which supports via API deletion; Clean up for OCI registry repo must be added before this test.")
def test_spack_create_cache_installation(tmp_path):
    install_dir = tmp_path
    spack_operation, env_path = setup(install_dir)
    assert isinstance(spack_operation, SpackOperationCreateCache)
    spack_operation.install_spack(bashrc_path=str(install_dir / Path('.bashrc')))
    spack_operation.setup_spack_env()
    spack_operation.concretize_spack_env()
    assert len(spack_operation.cache_dependency.list_tags()) > 0
    spack_operation.install_packages()
    assert len(spack_operation.build_cache.list_tags()) > 0
    assert spack_operation.check_installed_spack_packages(env_path) == True



# Dedal library - Wrapper over Spack for building multiple target
# environments: ESD, Virtual Boxes, HPC compatible kernels, etc.

#  (c) Copyright 2025 Dedal developers

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

from dedal.build_cache.BuildCacheManagerOci import BuildCacheManagerOci
from dedal.build_cache.BuildCacheManagerOciDefault import BuildCacheManagerOciDefault
from dedal.configuration.SpackConfig import SpackConfig
from dedal.enum.SpackCacheStorageEnum import SpackCacheStorageEnum


class BuildCacheManagerCreator:
    @staticmethod
    def get_build_cache_manager_operator(spack_config: SpackConfig = SpackConfig()):
        if spack_config.spack_cache_storage_type is SpackCacheStorageEnum.OCI:
            cache_dependency = BuildCacheManagerOci(os.environ.get('CONCRETIZE_OCI_HOST'),
                                                    os.environ.get('CONCRETIZE_OCI_PROJECT'),
                                                    os.environ.get('CONCRETIZE_OCI_USERNAME'),
                                                    os.environ.get('CONCRETIZE_OCI_PASSWORD'),
                                                    cache_version=spack_config.cache_version_concretize)
            build_cache = BuildCacheManagerOci(os.environ.get('BUILDCACHE_OCI_HOST'),
                                               os.environ.get('BUILDCACHE_OCI_PROJECT'),
                                               os.environ.get('BUILDCACHE_OCI_USERNAME'),
                                               os.environ.get('BUILDCACHE_OCI_PASSWORD'),
                                               cache_version=spack_config.cache_version_build)
            return cache_dependency, build_cache
        elif spack_config.spack_cache_storage_type is SpackCacheStorageEnum.OCI_DEFAULT:
            cache_dependency = BuildCacheManagerOciDefault(os.environ.get('CONCRETIZE_OCI_HOST'),
                                                           os.environ.get('CONCRETIZE_OCI_PROJECT'),
                                                           os.environ.get('CONCRETIZE_OCI_USERNAME'),
                                                           os.environ.get('CONCRETIZE_OCI_PASSWORD'),
                                                           cache_version=spack_config.cache_version_concretize)
            build_cache = BuildCacheManagerOciDefault(os.environ.get('BUILDCACHE_OCI_HOST'),
                                                      os.environ.get('BUILDCACHE_OCI_PROJECT'),
                                                      os.environ.get('BUILDCACHE_OCI_USERNAME'),
                                                      os.environ.get('BUILDCACHE_OCI_PASSWORD'),
                                                      cache_version=spack_config.cache_version_build)
            return cache_dependency, build_cache
        return None, None
